"""SC Controller - App

Main application window
"""

import json
import logging
import os
import platform
import re
import sys
import urllib

from gi.repository import Gdk, Gio, GLib, Gtk

from scc.actions import NoAction
from scc.config import Config
from scc.constants import DAEMON_VERSION, LEFT, RIGHT, STICK, STICK_PAD_MAX, SCButtons
from scc.custom import load_custom_module
from scc.gui.binding_editor import BindingEditor
from scc.gui.controller_image import ControllerImage
from scc.gui.controller_widget import BUTTONS, GYROS, PADS, STICKS, TRIGGERS
from scc.gui.daemon_manager import ControllerManager, DaemonManager
from scc.gui.dwsnc import IS_UNITY, headerbar
from scc.gui.parser import GuiActionParser, InvalidAction
from scc.gui.profile_switcher import ProfileSwitcher
from scc.gui.ribar import RIBar
from scc.gui.statusicon import get_status_icon
from scc.gui.userdata_manager import UserDataManager
from scc.modifiers import NameModifier
from scc.paths import get_config_path, get_profiles_path
from scc.profile import Profile
from scc.tools import (
	_,
	check_access,
	find_gksudo,
	find_profile,
	get_profile_name,
	nameof,
	profile_is_default,
	profile_is_override,
	set_logging_level,
)

log = logging.getLogger("App")


class App(Gtk.Application, UserDataManager, BindingEditor):
	"""Main application / window."""

	HILIGHT_COLOR = "#FF00FF00"  # ARGB
	OBSERVE_COLOR = "#FF60A0FF"  # ARGB
	CONFIG = "scc.config.json"
	RELEASE_URL = "https://github.com/C0rn3j/sc-controller/releases/tag/v%s"
	OSD_MODE_PROF_NAME = ".scc-osd.profile_editor"

	def __init__(self, gladepath: str = "/usr/share/scc", imagepath: str = "/usr/share/scc/images"):
		Gtk.Application.__init__(
			self,
			application_id="me.c0rn3j.scc",
			flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE | Gio.ApplicationFlags.NON_UNIQUE,
		)
		UserDataManager.__init__(self)
		BindingEditor.__init__(self, self)
		# Setup Gtk.Application
		self.convert_old_profiles()
		self.setup_commandline()
		# Setup DaemonManager
		self.dm = DaemonManager()
		self.dm.connect("alive", self.on_daemon_alive)
		self.dm.connect("event", self.on_daemon_event_observer)
		self.dm.connect("controller-count-changed", self.on_daemon_ccunt_changed)
		self.dm.connect("dead", self.on_daemon_dead)
		self.dm.connect("error", self.on_daemon_error)
		(self.dm.connect("reconfigured", self.on_daemon_reconfigured),)
		self.dm.connect("version", self.on_daemon_version)
		# Load custom stuff
		load_custom_module(log, "gui")
		# Set variables
		self.config = Config()
		self.gladepath = gladepath
		self.imagepath = imagepath
		self.builder = None
		self.recursing = False
		self.statusicon = None
		self.status = "unknown"
		self.context_menu_for = None
		self.daemon_changed_profile = False
		self.osd_mode = False  # In OSD mode, only active profile can be editted
		self.osd_mode_mapper = None
		self.background = None
		self.outdated_version = None
		self.profile_switchers = []
		self.test_mode_controller = None
		self.current_ui_layout = "default"  # only "default" and "deck" are supported
		self.current_file = None  # Currently edited file
		self.controller_count = 0
		self.current = Profile(GuiActionParser())
		self.just_started = True
		self.button_widgets = {}
		self.hilights = {App.HILIGHT_COLOR: set(), App.OBSERVE_COLOR: set()}
		self.undo = []
		self.redo = []

	def setup_widgets(self):
		# Important stuff
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(self.gladepath, "app.glade"))
		self.builder.connect_signals(self)
		self.window = self.builder.get_object("window")
		self.add_window(self.window)
		self.window.set_title(_("SC Controller"))
		self.window.set_wmclass("SC Controller", "SC Controller")
		self.ribar = None
		self.create_binding_buttons()

		ps = self.add_switcher(12, 12)
		ps.set_allow_new(True)
		ps.set_profile(self.load_profile_selection())
		ps.connect("new-clicked", self.on_new_clicked)
		ps.connect("save-clicked", self.on_save_clicked)

		# Drag&drop target
		self.builder.get_object("content").drag_dest_set(
			Gtk.DestDefaults.ALL,
			[
				Gtk.TargetEntry.new("text/uri-list", Gtk.TargetFlags.OTHER_APP, 0),
				Gtk.TargetEntry.new("text/plain", Gtk.TargetFlags.OTHER_APP, 0),
			],
			Gdk.DragAction.COPY,
		)

		# 'C' and 'CPAD' buttons
		vbc = self.builder.get_object("vbC")
		self.main_area = self.builder.get_object("mainArea")
		vbc.get_parent().remove(vbc)

		# Background
		self.background = ControllerImage(self)
		self.background.connect("hover", self.on_background_area_hover)
		self.background.connect("leave", self.on_background_area_hover, None)
		self.background.connect("click", self.on_background_area_click)
		self.background.connect("button-press-event", self.on_background_button_press)
		self.main_area.put(self.background, 0, 0)
		self.main_area.put(vbc, 0, 0)  # (self.IMAGE_SIZE[0] / 2) - 90, self.IMAGE_SIZE[1] - 100)

		# Test markers (those blue circles over PADs and sticks)
		self.lpad_test = Gtk.Image.new_from_file(os.path.join(self.imagepath, "test-cursor.svg"))
		self.rpad_test = Gtk.Image.new_from_file(os.path.join(self.imagepath, "test-cursor.svg"))
		self.stick_test = Gtk.Image.new_from_file(os.path.join(self.imagepath, "test-cursor.svg"))
		self.main_area.put(self.lpad_test, 40, 40)
		self.main_area.put(self.rpad_test, 290, 90)
		self.main_area.put(self.stick_test, 150, 40)

		# OSD mode (if used)
		if self.osd_mode:
			self.builder.get_object("btDaemon").set_sensitive(False)
			self.window.set_title(_("Edit Profile"))

		# Headerbar
		headerbar(self.builder.get_object("hbWindow"))

	def load_gui_config_for_controller(self, controller, first):
		"""Loads controller config, changes image and hides, shows or disables
		buttons around it.

		To make this look less jumpy, Gtk.Stack is used to make transition
		to empty page and only after that is grid repopulated, everything
		set up and Stack switched back to original page.
		"""
		stckEditor = self.builder.get_object("stckEditor")
		lblEmpty = self.builder.get_object("lblEmpty")
		if controller:
			config = controller.load_gui_config(self.imagepath or {})
		else:
			config = {}
		config = self.background.use_config(config, controller=controller)

		def do_loading():
			"""Called after transition is finished"""
			self.background.use_config(config, controller=controller)
			self.apply_gui_config_buttons(config)

		if first:
			b1 = self.background.get_config()["gui"]["background"]
			b2 = config["gui"]["background"]
			if b1 == b2:
				# If application has just started and image is
				# not changing, transition would just look weird
				do_loading()
				return
		if not first:
			stckEditor.set_transition_type(Gtk.StackTransitionType.SLIDE_DOWN)
		stckEditor.set_visible_child(lblEmpty)
		GLib.timeout_add(stckEditor.get_transition_duration(), do_loading)

	def apply_gui_config_buttons(self, config):
		"""Changes UI according to controller configuration"""
		stckEditor = self.builder.get_object("stckEditor")
		grEditor = self.builder.get_object("grEditor")
		btCPAD = self.builder.get_object("btCPAD")
		btDPAD = self.builder.get_object("btDPAD")
		btGYRO = self.builder.get_object("btGYRO")
		btC = self.builder.get_object("btC")

		buttons = ControllerImage.get_names(config.get("buttons", {}))
		axes = ControllerImage.get_names(config.get("axes", {}))
		gyros = config.get("gyros", False)
		# Set sensitivity to signalize available inputs
		# Buttons (as on image)
		for b in BUTTONS:
			w = self.builder.get_object("bt" + nameof(b))
			if w:
				w.set_sensitive(nameof(b) in buttons)
		# Buttons (as GTK Widgets)
		for b in self.button_widgets:
			try:
				w = self.button_widgets[b]
				icon, trash = ControllerManager.get_button_icon(config, b, True)
				w.icon.set_from_file(icon)
			except Exception:
				pass
		# Triggers
		w = self.builder.get_object("btLT")
		if w:
			w.set_sensitive("ltrig" in axes)
		w = self.builder.get_object("btRT")
		if w:
			w.set_sensitive("rtrig" in axes)
		# Sticks & pads
		for b in PADS + STICKS:
			w = self.builder.get_object("bt" + nameof(b))
			if w:
				w.set_sensitive(b.lower() + "_x" in axes or b.lower() + "_y" in axes or nameof(b) in buttons)
		# Gyro
		for b in GYROS:
			w = self.builder.get_object("bt" + b)
			if w:
				# TODO: Maybe actual detection
				w.set_sensitive(gyros)

		for w in (btC, btCPAD, btDPAD, btGYRO):
			w.set_visible(w.get_sensitive())

		# Re-layout if needed
		expected_layout = "default"
		if len(axes) >= 8 and btC.get_sensitive():
			expected_layout = "deck"

		if expected_layout != self.current_ui_layout:
			self.apply_ui_layout(expected_layout)

		stckEditor.set_visible_child(grEditor)
		GLib.idle_add(self.on_c_size_allocate)

	def apply_ui_layout(self, layout):
		"""Changes layout of ui elements to fit additional buttons needed for Deck
		"""
		if layout == "deck":
			# Move 'C' button bellow LGRIP
			btRGRIP = self.builder.get_object("btRGRIP")
			btC = self.builder.get_object("btC")
			btC.get_parent().remove(btC)
			btC.set_margin_right(0)
			btRGRIP.get_parent().pack_start(btC, False, True, 0)
			btRGRIP.get_parent().reorder_child(btC, 5)
			# Move 'GYRO' button to middle of image (where C was)
			btGYRO = self.builder.get_object("btGYRO")
			btGYRO.get_parent().remove(btGYRO)
			vbC = self.builder.get_object("vbC")
			vbC.pack_start(btGYRO, False, True, 0)
			btGYRO.set_margin_top(30)
			# Resize buttons at bottom
			# for w in ['btSTICK', 'btRSTICK', 'btLPAD', 'btRPAD']:
			# w.set_size_request(150, -1)
			# Move 'DPAD' bellow 'LGRIP'
			btLGRIP = self.builder.get_object("btLGRIP")
			btDPAD = self.builder.get_object("btDPAD")
			btDPAD.get_parent().remove(btDPAD)
			btLGRIP.get_parent().pack_start(btDPAD, False, True, 6)
			btLGRIP.get_parent().reorder_child(btDPAD, 5)

	def setup_statusicon(self) -> None:
		menu = self.builder.get_object("mnuTray")
		self.statusicon = get_status_icon(self.imagepath, menu)
		self.statusicon.connect("clicked", self.on_statusicon_clicked)
		# if not self.statusicon.is_clickable():
		# self.builder.get_object("mnuShowWindowTray").set_visible(True)
		# Workaround - always add it to the menu, see https://github.com/C0rn3j/sc-controller/issues/53
		self.builder.get_object("mnuShowWindowTray").set_visible(True)
		GLib.idle_add(self.statusicon.set, f"scc-{self.status}", _("SC Controller"))

	def destroy_statusicon(self):
		self.statusicon.destroy()
		self.statusicon = None

	def check(self):
		"""Performs various (three) checks and reports possible problems"""
		# TODO: Maybe not best place to do this
		try:
			# Dynamic modules
			with open("/proc/modules") as file:
				rawlist = file.read().split("\n")
			kernel_mods = [line.split(" ")[0] for line in rawlist]
			# Built-in modules
			release = platform.uname()[2]
			with open("/lib/modules/%s/modules.builtin" % release) as file:
				rawlist = file.read().split("\n")
			kernel_mods += [os.path.split(x)[-1].split(".")[0] for x in rawlist]
		except Exception:
			# Maybe running on BSD or Windows...
			kernel_mods = []

		if len(kernel_mods) > 0 and "uinput" not in kernel_mods:
			# There is no uinput
			msg = _("uinput kernel module not loaded")
			msg += "\n\n" + _("Please, consult your distribution manual on how to enable uinput")
			msg += "\n" + _('or click on "Fix Temporary" button to attempt fix that should work until next restart.')
			ribar = self.show_error(msg)
			gksudo = find_gksudo()
			if gksudo and not hasattr(ribar, "_fix_tmp"):
				button = Gtk.Button.new_with_label(_("Fix Temporary"))
				ribar._fix_tmp = button
				button.connect(
					"clicked",
					self.apply_temporary_fix,
					gksudo + ["modprobe", "uinput"],
					_("This will load missing uinput module."),
				)
				ribar.add_button(button, -1)
			return True
		if not os.path.exists("/dev/uinput"):
			# /dev/uinput missing
			msg = _("/dev/uinput doesn't exists")
			msg += "\n" + _("uinput kernel module is loaded, but /dev/uinput is missing.")
			# msg += "\n\n" + _('Please, consult your distribution manual on what in the world could cause this.')
			msg += "\n\n" + _("Please, consult your distribution manual on how to enable uinput")
			self.show_error(msg)
			return True
		if not check_access("/dev/uinput"):
			# Cannot acces uinput
			msg = _("You don't have required access to /dev/uinput.")
			msg += "\n" + _("This will most likely prevent emulation from working.")
			msg += "\n\n" + _("Please, consult your distribution manual on how to enable uinput")
			msg += "\n" + _('or click on "Fix Temporary" button to attempt fix that should work until next restart.')
			ribar = self.show_error(msg)
			gksudo = find_gksudo()
			if gksudo and not hasattr(ribar, "_fix_tmp"):
				button = Gtk.Button.new_with_label(_("Fix Temporary"))
				ribar._fix_tmp = button
				button.connect(
					"clicked",
					self.apply_temporary_fix,
					gksudo + ["chmod", "666", "/dev/uinput"],
					_(
						"This will enable input emulation for <i>every application</i> and <i>all users</i> on this machine.",
					),
				)
				ribar.add_button(button, -1)
			return True
		return False

	def apply_temporary_fix(self, trash, shell_command, message):
		"""Display MessageBox with confirmation, try to run passed shell command and restart daemon.

		Doing this allows user to teporary fix some uinput-related problems
		by his vaim belief I'll not format his harddrive.
		"""
		d = Gtk.MessageDialog(
			parent=self.window,
			flags=Gtk.DialogFlags.MODAL,
			type=Gtk.MessageType.WARNING,
			buttons=Gtk.ButtonsType.OK_CANCEL,
			message_format=_("sudo fix-my-pc"),
		)

		def on_response(dialog, response_id):
			if response_id == -5:  # OK button, not defined anywhere
				sudo = Gio.Subprocess.new(shell_command, 0)
				sudo.communicate(None, None)
				if sudo.get_exit_status() == 0:
					self.dm.restart()
				else:
					d2 = Gtk.MessageDialog(
						parent=d,
						flags=Gtk.DialogFlags.MODAL,
						type=Gtk.MessageType.ERROR,
						buttons=Gtk.ButtonsType.OK,
						message_format=_("Command Failed"),
					)
					d2.run()
					d2.destroy()
			d.destroy()

		d.connect("response", on_response)
		d.format_secondary_markup(
			_("""Following command is going to be executed:

<b>%s</b>

%s""")
			% (" ".join(shell_command), message),
		)
		d.show()

	def hilight(self, button):
		"""Hilights specified button on background image"""
		if button:
			self.hilights[App.HILIGHT_COLOR] = set([button])
		else:
			self.hilights[App.HILIGHT_COLOR] = set()
		self._update_background()

	def _update_background(self):
		h = {}
		for color in self.hilights:
			for i in self.hilights[color]:
				h[i] = color
		self.background.hilight(h)

	def hint(self, button):
		"""As hilight, but marks GTK Button as well"""
		active = None
		for b in self.button_widgets.values():
			if b.widget.get_sensitive():
				b.widget.set_state(Gtk.StateType.NORMAL)
				if b.name == button:
					active = b.widget

		if active is not None:
			active.set_state(Gtk.StateType.ACTIVE)

		self.hilight(button)

	def show_editor(self, id):
		action = self.get_action(self.current, id)
		ae = self.choose_editor(action, "", id)
		ae.allow_first_page()
		ae.set_input(id, action)
		ae.show(self.window)

	def show_context_menu(self, for_id):
		"""Sets sensitivity of popup menu items and displays it on screen"""
		mnuPopup = self.builder.get_object("mnuPopup")
		mnuCopy = self.builder.get_object("mnuCopy")
		mnuClear = self.builder.get_object("mnuClear")
		mnuPaste = self.builder.get_object("mnuPaste")
		mnuEPress = self.builder.get_object("mnuEditPress")
		mnuEPressS = self.builder.get_object("mnuEditPressSeparator")
		self.context_menu_for = for_id
		clp = Gtk.Clipboard.get_default(Gdk.Display.get_default())
		mnuCopy.set_sensitive(bool(self.get_action(self.current, for_id)))
		mnuClear.set_sensitive(bool(self.get_action(self.current, for_id)))
		mnuPaste.set_sensitive(clp.wait_is_text_available())
		mnuEPress.set_visible(for_id in STICKS + PADS)
		mnuEPressS.set_visible(mnuEPress.get_visible())

		mnuPopup.popup(None, None, None, None, 3, Gtk.get_current_event_time())

	def save_config(self):
		self.config.save()
		self.dm.reconfigure()
		self.enable_test_mode()

	def on_statusicon_clicked(self, *a) -> None:
		"""Handler for user clicking on tray icon button."""
		self.window.set_visible(not self.window.get_visible())

	def on_window_delete_event(self, *a):
		"""Called when user tries to close window"""
		if not IS_UNITY and self.config["gui"]["enable_status_icon"] and self.config["gui"]["minimize_to_status_icon"]:
			# Override closing and hide instead
			self.window.set_visible(False)
		else:
			self.on_mnuExit_activate()
		return True

	def on_mnuClear_activate(self, *a):
		"""Handler for 'Clear' context menu item.
		Simply sets NoAction to input.
		"""
		self.on_action_chosen(self.context_menu_for, NoAction())

	def on_mnuCopy_activate(self, *a):
		"""Handler for 'Copy' context menu item.
		Converts action to string and sends that string to clipboard.
		"""
		a = self.get_action(self.current, self.context_menu_for)
		if a:
			if a.name:
				a = NameModifier(a.name, a)
			clp = Gtk.Clipboard.get_default(Gdk.Display.get_default())
			clp.set_text(a.to_string().encode("utf-8"), -1)
			clp.store()

	def on_mnuPaste_activate(self, *a):
		"""Handler for 'Paste' context menu item.
		Reads string from clipboard, parses it as action and sets that action
		on selected input.
		"""
		clp = Gtk.Clipboard.get_default(Gdk.Display.get_default())
		text = clp.wait_for_text()
		if text:
			a = GuiActionParser().restart(text.decode("utf-8")).parse()
			if not isinstance(a, InvalidAction):
				self.on_action_chosen(self.context_menu_for, a)

	def on_mnuEditPress_activate(self, *a):
		"""Handler for 'Edit Pressed Action' context menu item.
		"""
		id = self.context_menu_for
		if id == STICK:
			id = nameof(SCButtons.STICKPRESS)
		self.show_editor(getattr(SCButtons, id))

	def on_mnuGlobalSettings_activate(self, *a):
		from scc.gui.global_settings import GlobalSettings

		gs = GlobalSettings(self)
		gs.show(self.window)

	def on_mnuImport_activate(self, *a):
		"""Handler for 'Import Steam Profile' context menu item.
		Displays apropriate dialog.
		"""
		from scc.gui.importexport.dialog import Dialog

		ied = Dialog(self)
		ied.show(self.window)

	def on_btUndo_clicked(self, *a):
		if len(self.undo) < 1:
			return
		undo, self.undo = self.undo[-1], self.undo[0:-1]
		self.set_action(self.current, undo.id, undo.before)
		self.redo.append(undo)
		self.builder.get_object("btRedo").set_sensitive(True)
		if len(self.undo) < 1:
			self.builder.get_object("btUndo").set_sensitive(False)
		self.on_profile_modified()

	def on_btRedo_clicked(self, *a):
		if len(self.redo) < 1:
			return
		redo, self.redo = self.redo[-1], self.redo[0:-1]
		self.set_action(self.current, redo.id, redo.after)
		self.undo.append(redo)
		self.builder.get_object("btUndo").set_sensitive(True)
		if len(self.redo) < 1:
			self.builder.get_object("btRedo").set_sensitive(False)
		self.on_profile_modified()

	def on_profiles_loaded(self, profiles):
		for ps in self.profile_switchers:
			ps.set_profile_list(profiles)

	def undeletable_dialog(self, dlg, *a):
		dlg.hide()
		return True

	def on_btNewProfile_clicked(self, *a):
		"""Called when new profile name is set and OK is clicked."""
		txNewProfile = self.builder.get_object("txNewProfile")
		rbNewProfile = self.builder.get_object("rbNewProfile")

		dlg = self.builder.get_object("dlgNewProfile")
		if rbNewProfile.get_active():
			# Creating blank profile is requested
			self.current.clear()
		else:
			self.current.is_template = False
		self.new_profile(self.current, txNewProfile.get_text())
		dlg.hide()

	def on_rbNewProfile_group_changed(self, *a):
		"""Called when user clicks 'Copy current profile' button.

		If profile name was not changed by user before clicking it,
		it's automatically changed.
		"""
		txNewProfile = self.builder.get_object("txNewProfile")
		rbNewProfile = self.builder.get_object("rbNewProfile")

		if not txNewProfile._changed:
			self.recursing = True
			if rbNewProfile.get_active():
				# Create empty profile
				txNewProfile.set_text(self.generate_new_name())
			else:
				# Copy current profile
				txNewProfile.set_text(self.generate_copy_name(txNewProfile._name))
			self.recursing = False

	def on_profile_modified(self, update_ui: bool = True):
		"""Called when selected profile is modified in memory."""
		if update_ui:
			self.profile_switchers[0].set_profile_modified(True, self.current.is_template)

		if not self.current_file.get_path().endswith(".mod"):
			mod = self.current_file.get_path() + ".mod"
			self.current_file = Gio.File.new_for_path(mod)

		self.save_profile(self.current_file, self.current)

	def on_profile_loaded(self, profile: Profile, giofile: Gio.File):
		self.current = profile
		self.current_file = giofile
		self.recursing = True
		self.profile_switchers[0].set_profile_modified(False, self.current.is_template)
		self.builder.get_object("txProfileFilename").set_text(giofile.get_path())
		self.builder.get_object("txProfileDescription").get_buffer().set_text(self.current.description)
		self.builder.get_object("cbProfileIsTemplate").set_active(self.current.is_template)
		for b in self.button_widgets.values():
			b.update()
		self.recursing = False

	def on_profile_selected(self, ps, name, giofile: Gio.File):
		if ps == self.profile_switchers[0]:
			self.load_profile(giofile)
		if ps.get_controller():
			ps.get_controller().set_profile(giofile.get_path())

	def on_unknown_profile(self, ps, name):
		log.warning("Daemon reported unknown profile: '%s'; Overriding.", name)
		if self.current_file is not None and ps.get_controller() is not None:
			ps.get_controller().set_profile(self.current_file.get_path())

	def on_save_clicked(self, *a):
		if self.current_file.get_path().endswith(".mod"):
			orig = self.current_file.get_path()[0:-4]
			self.current_file = Gio.File.new_for_path(orig)

		if self.current.is_template:
			# Ask user if he is OK with overwriting template
			d = Gtk.MessageDialog(
				parent=self.window,
				flags=Gtk.DialogFlags.MODAL,
				type=Gtk.MessageType.QUESTION,
				buttons=Gtk.ButtonsType.YES_NO,
				message_format=_("You are about to save changes over template.\nAre you sure?"),
			)
			NEW_PROFILE_BUTTON = 7
			d.add_button(_("Create New Profile"), NEW_PROFILE_BUTTON)

			r = d.run()
			d.destroy()
			if r == NEW_PROFILE_BUTTON:
				# New profile button clicked
				ps = self.profile_switchers[0]
				rbCopyProfile = self.builder.get_object("rbCopyProfile")
				self.on_new_clicked(ps, ps.get_profile_name())
				rbCopyProfile.set_active(True)
				return
			if r != -8:
				# Bail out if user answers anything but yes
				return

		self.save_profile(self.current_file, self.current)

	def on_switch_to_clicked(self, ps, *a):
		"""Switches editor to another controller"""
		ps0 = self.profile_switchers[0]
		if ps == ps0:
			return

		c, p = ps.get_controller(), ps.get_profile_name()
		c0, p0 = ps0.get_controller(), ps0.get_profile_name()

		ps0.set_controller(c)
		ps0.set_profile(p)
		ps.set_controller(c0)
		ps.set_profile(p0)

		self.load_gui_config_for_controller(c, False)
		self.enable_test_mode()

	def on_profile_saved(self, giofile: Gio.File, send: bool = True):
		"""Called when selected profile is saved to disk
		"""
		if self.osd_mode:
			# Special case, profile shouldn't be changed while in osd_mode
			if not giofile.get_path().endswith(".mod"):
				self.profile_switchers[0].set_profile_modified(False, self.current.is_template)
			return

		if giofile.get_path().endswith(".mod"):
			# Special case, this one is saved only to be sent to daemon
			# and user doesn't need to know about it
			if self.dm.is_alive():
				controller = self.profile_switchers[0].get_controller()
				if controller:
					controller.set_profile(giofile.get_path())
				else:
					self.dm.set_profile(giofile.get_path())
			return

		self.profile_switchers[0].set_profile_modified(False, self.current.is_template)
		if send and self.dm.is_alive() and not self.daemon_changed_profile:
			for ps in self.profile_switchers:
				controller = ps.get_controller()
				if controller:
					active = controller.get_profile()
					if active.endswith(".mod"):
						active = active[0:-4]
					if active == giofile.get_path():
						controller.set_profile(giofile.get_path())

		self.current_file = giofile

	def generate_new_name(self):
		"""Generates name for new profile.
		That is 'New Profile X', where X is number that makes name unique.
		"""
		i = 1
		new_name = _("New Profile %s") % (i,)
		filename = os.path.join(get_profiles_path(), new_name + ".sccprofile")
		while os.path.exists(filename):
			i += 1
			new_name = _("New Profile %s") % (i,)
			filename = os.path.join(get_profiles_path(), new_name + ".sccprofile")
		return new_name

	def generate_copy_name(self, name):
		"""Generates name for profile copy.
		That is 'New Profile X', where X is number that makes name unique.
		"""
		new_name = _("%s (copy)") % (name,)
		filename = os.path.join(get_profiles_path(), new_name + ".sccprofile")
		i = 2
		while os.path.exists(filename):
			new_name = _("%s (copy %s)") % (name,)
			filename = os.path.join(get_profiles_path(), new_name + ".sccprofile")
			i += 1
		return new_name

	def on_txNewProfile_changed(self, tx):
		if self.recursing:
			return
		tx._changed = True

	def on_new_clicked(self, ps, name):
		dlg = self.builder.get_object("dlgNewProfile")
		txNewProfile = self.builder.get_object("txNewProfile")
		rbNewProfile = self.builder.get_object("rbNewProfile")
		self.recursing = True
		rbNewProfile.set_active(True)
		txNewProfile.set_text(self.generate_new_name())
		txNewProfile._name = name
		txNewProfile._changed = False
		self.recursing = False
		dlg.set_transient_for(self.window)
		dlg.show()

	def on_action_chosen(self, id, action, mark_changed=True):
		before = self.set_action(self.current, id, action)
		if mark_changed:
			if before.to_string() != action.to_string():
				# TODO: Maybe better comparison
				self.undo.append(UndoRedo(id, before, action))
				self.builder.get_object("btUndo").set_sensitive(True)
			self.on_profile_modified()
		else:
			self.on_profile_modified(update_ui=False)
		return before

	def on_background_area_hover(self, trash, area):
		self.hint(area)

	def on_background_button_press(self, trash, event):
		if event.button == 3:
			mnuImage = self.builder.get_object("mnuImage")
			mnuImage.popup(None, None, None, None, 3, Gtk.get_current_event_time())

	def on_mnu_change_background_image(self, mnu, *a):
		command, filename = mnu.get_name().split(",")
		if command == "background":
			self.background.override_background(filename)
		elif command == "buttons":
			self.background.override_buttons(filename)
			self.apply_gui_config_buttons(self.background.get_config())
		elif command == "undo":
			self.background.undo_override()
			self.apply_gui_config_buttons(self.background.get_config())

	def on_background_area_click(self, trash, area):
		if area in [x.name for x in BUTTONS]:
			self.hint(None)
			self.show_editor(getattr(SCButtons, area))
		elif area in TRIGGERS + STICKS + PADS:
			self.hint(None)
			self.show_editor(area)

	def on_c_size_allocate(self, *a):
		"""Called when size of 'Button C' or CPAD is changed.
		Centers buttons on background image
		"""
		main_area = self.builder.get_object("mainArea")
		y = main_area.get_allocation().height - 5
		w = self.builder.get_object("vbC")
		allocation = w.get_allocation()
		x = (self.background.get_allocation().width - allocation.width) / 2
		y -= allocation.height

		if self.background.get_config()["gui"]["no_buttons_in_gui"]:
			# no_buttons_in_gui is used to keep image without changes
			# This moves "C" button away so it doesn't obscure it as well
			y = 10

		if w.get_parent():
			main_area.move(w, x, y)
		else:
			main_area.put(w, x, y)
		return False

	def on_ebImage_motion_notify_event(self, box, event):
		self.background.on_mouse_moved(event.x, event.y)

	def on_exiting_n_daemon_killed(self, *a):
		self.quit()

	def on_mnuExit_activate(self, *a):
		if not self.osd_mode and self.app.config["gui"]["autokill_daemon"]:
			log.debug("Terminating scc-daemon")
			for x in ("content", "mnuEmulationEnabled", "mnuEmulationEnabledTray"):
				w = self.builder.get_object(x)
				w.set_sensitive(False)
			self.set_daemon_status("unknown", False)
			self.hide_error()
			if self.dm.is_alive():
				self.dm.connect("dead", self.on_exiting_n_daemon_killed)
				self.dm.connect("error", self.on_exiting_n_daemon_killed)
				self.dm.stop()
			else:
				# Daemon appears to be dead, kill it just in case
				self.dm.stop()
				self.quit()
		else:
			self.quit()

	def on_mnuAbout_activate(self, *a) -> None:
		from scc.gui.aboutdialog import AboutDialog

		AboutDialog(self).show(self.window)

	def on_daemon_alive(self, *a):
		self.set_daemon_status("alive", True)
		if not self.release_notes_visible():
			self.hide_error()
		self.just_started = False
		if self.osd_mode:
			self.enable_osd_mode()
		elif self.profile_switchers[0].get_file() is not None and not self.just_started:
			self.dm.set_profile(self.current_file.get_path())
		GLib.timeout_add_seconds(1, self.check)
		self.enable_test_mode()

	def on_daemon_ccunt_changed(self, daemon, count):
		if self.controller_count == 0:
			# First controller connected
			#
			# 'event' signal should be connected only on first controller,
			# so this block is executed only when number of connected
			# controllers changes from 0 to 1
			if len(self.dm.get_controllers()) > 0:
				c = self.dm.get_controllers()[0]
				self.load_gui_config_for_controller(c, first=True)
		if count > self.controller_count:
			# Controller added
			while len(self.profile_switchers) < count:
				s = self.add_switcher()
		elif count < self.controller_count:
			# Controller removed
			while len(self.profile_switchers) > max(1, count):
				s = self.profile_switchers.pop()
				s.set_controller(None)
				self.remove_switcher(s)

		# Assign controllers to widgets
		for i in range(count):
			c = self.dm.get_controllers()[i]
			self.profile_switchers[i].set_controller(c)

		if count < 1:
			# Special case, no controllers are connected, but one widget
			# has to stay on screen
			self.profile_switchers[0].set_controller(None)
			self.load_gui_config_for_controller(None, first=True)

		self.controller_count = count

	def new_profile(self, profile: Profile, name: str):
		filename = os.path.join(get_profiles_path(), name + ".sccprofile")
		self.current_file = Gio.File.new_for_path(filename)
		self.save_profile(self.current_file, profile)
		controller = self.profile_switchers[0].get_controller()
		if controller:
			controller.set_profile(filename)
		else:
			self.dm.set_profile(filename)
		self.profile_switchers[0].set_profile(name, create=True)

	def add_switcher(self, margin_left=24, margin_right=24):
		"""Adds new profile switcher widgets on top of window. Called
		when new controller is connected to daemon.

		Returns generated ProfileSwitcher instance.
		"""
		vbSwitchers = self.builder.get_object("vbSwitchers")
		sepSwitchers = self.builder.get_object("sepSwitchers")

		ps = ProfileSwitcher(self.imagepath, self.config)
		ps.set_margin_left(margin_left)
		ps.set_margin_right(margin_right)
		ps.connect("right-clicked", self.on_profile_right_clicked)
		ps.connect("switch-to-clicked", self.on_switch_to_clicked)

		vbSwitchers.pack_start(ps, False, False, 0)
		vbSwitchers.reorder_child(ps, 0)
		if len(vbSwitchers.get_children()) == 2:
			# 1st switcher is bellow separator, rest is stacked on top.
			# That means separator should be moved and shown when 2nd
			# switcher is created.
			vbSwitchers.reorder_child(sepSwitchers, 0)
			sepSwitchers.set_visible(True)
		vbSwitchers.show_all()

		if self.osd_mode:
			ps.set_allow_switch(False)

		if len(self.profile_switchers) > 0:
			ps.set_profile_list(self.profile_switchers[0].get_profile_list())
			ps.set_switch_to_enabled(True)

		self.profile_switchers.append(ps)
		ps.connect("changed", self.on_profile_selected)
		ps.connect("unknown-profile", self.on_unknown_profile)
		return ps

	def remove_switcher(self, s):
		"""Removes given profile switcher from UI.
		"""
		vbSwitchers = self.builder.get_object("vbSwitchers")
		sepSwitchers = self.builder.get_object("sepSwitchers")
		vbSwitchers.remove(s)
		s.destroy()
		if len(vbSwitchers.get_children()) == 2:
			sepSwitchers.set_visible(False)

	def enable_test_mode(self):
		"""Disables and re-enables Input Test mode. If sniffing is disabled in
		daemon configuration, 2nd call fails and logs error.
		"""
		if self.dm.is_alive() and not self.osd_mode:
			if self.test_mode_controller:
				self.test_mode_controller.unlock_all()
			try:
				c = self.dm.get_controllers()[0]
			except IndexError:
				# Zero controllers
				return
			if c:
				c.unlock_all()
				c.observe(
					DaemonManager.nocallback,
					self.on_observe_failed,
					"A",
					"B",
					"C",
					"X",
					"Y",
					"START",
					"BACK",
					"LB",
					"RB",
					"LPAD",
					"RPAD",
					"LGRIP",
					"RGRIP",
					"LT",
					"RT",
					"LEFT",
					"RIGHT",
					"STICK",
					"STICKPRESS",
				)
				self.test_mode_controller = c

	def enable_osd_mode(self):
		# TODO: Support for multiple controllers here
		self.osd_mode_controller = 0
		osd_mode_profile = Profile(GuiActionParser())
		osd_mode_profile.load(find_profile(App.OSD_MODE_PROF_NAME))
		try:
			c = self.dm.get_controllers()[self.osd_mode_controller]
		except IndexError:
			log.error("osd_mode: Controller not connected")
			self.quit()
			return

		def on_lock_failed(*a):
			log.error("osd_mode: Locking failed")
			self.quit()

		def on_lock_success(*a):
			log.debug("osd_mode: Locked everything")
			from scc.gui.osd_mode import OSDModeMapper, OSDModeMappings

			self.osd_mode_mapper = OSDModeMapper(self, osd_mode_profile)
			self.osd_mode_mapper.set_target_window(self.window.get_window())
			self.builder.get_object("btUndo").set_visible(False)
			self.builder.get_object("btRedo").set_visible(False)

			m = OSDModeMappings(self, self.osd_mode_mapper, self.builder.get_object("OsdmodeMappings"))
			m.set_controller(self.profile_switchers[0].get_controller())
			m.show()

		# Locks everything but pads. Pads are emulating mouse and this is
		# better left in daemon - involving socket in mouse controls
		# adds too much lags.
		c.lock(
			on_lock_success,
			on_lock_failed,
			"A",
			"B",
			"X",
			"Y",
			"START",
			"BACK",
			"LB",
			"RB",
			"C",
			"STICK",
			"LGRIP",
			"RGRIP",
			"LT",
			"RT",
			"STICKPRESS",
		)

		# Ask daemon to temporaly reconfigure pads for mouse emulation
		c.replace(DaemonManager.nocallback, on_lock_failed, LEFT, osd_mode_profile.pads[LEFT])
		c.replace(DaemonManager.nocallback, on_lock_failed, RIGHT, osd_mode_profile.pads[RIGHT])

	def on_observe_failed(self, error):
		log.debug("Failed to enable test mode: %s", error)

	def on_daemon_version(self, daemon, version):
		"""Checks if reported version matches expected one.
		If not, daemon is restarted.
		"""
		if version != DAEMON_VERSION and self.outdated_version != version:
			log.warning(
				"Running daemon instance is too old (version %s, expected %s). Restarting...", version, DAEMON_VERSION,
			)
			self.outdated_version = version
			self.set_daemon_status("unknown", False)
			self.dm.restart()
		# At this point, correct daemon version of daemon is running
		# and we can check if there is anything new to inform user about
		elif self.app.config["gui"]["news"]["last_version"] != App.get_release():
			if self.app.config["gui"]["news"]["enabled"]:
				if not self.osd_mode:
					self.check_release_notes()

	def on_daemon_error(self, daemon, error):
		log.debug("Daemon reported error '%s'", error)
		msg = _("There was an error with enabling emulation: <b>%s</b>") % (error,)
		# Known errors are handled with aditional message
		if "Device not found" in error:
			msg += "\n" + _("Please, check if you have receiver dongle connected to USB port.")
		elif "LIBUSB_ERROR_ACCESS" in error:
			msg += "\n" + _("You don't have access to controller device.")
			msg += "\n\n" + (
				_(
					"Consult your distribution manual, try installing Steam package or <a href='%s'>install required udev rules manually</a>.",
				)
				% "https://wiki.archlinux.org/index.php/Gamepad#Steam_Controller_not_pairing"
			)
			# TODO: Write howto somewhere instead of linking to ArchWiki
		elif "LIBUSB_ERROR_BUSY" in error:
			msg += "\n" + _("Another application (most likely Steam) is using the controller.")
		elif "CANT_SUMMON_THE_DAEMON" in error:
			msg += "\n" + _(
				'Background process responsible for emulation is not starting.\n\nTry executing "scc-daemon debug" in terminal window to check for any errors'
				"\nor <a href='https://github.com/C0rn3j/sc-controller/issues'>open issue on GitHub</a> and copy output there.",
			)
		elif "LIBUSB_ERROR_PIPE" in error:
			msg += "\n" + _("USB dongle was removed.")
		elif "Failed to create uinput device." in error:
			# Call check() method and try to determine what went wrong.
			if self.check():
				# Check() returns True if error was "handled".
				return
			# If check() fails to find error reason, error message is displayed as it is

		if self.osd_mode:
			self.quit()

		self.show_error(msg)
		self.set_daemon_status("error", True)

	def on_daemon_event_observer(self, daemon, c, what, data):
		if self.osd_mode_mapper:
			self.osd_mode_mapper.handle_event(daemon, what, data)
		elif what in (LEFT, RIGHT, STICK):
			widget, area = {
				LEFT: (self.lpad_test, "LPADTEST"),
				RIGHT: (self.rpad_test, "RPADTEST"),
				STICK: (self.stick_test, "STICKTEST"),
			}[what]
			# Check if stick or pad is released
			if data[0] == data[1] == 0:
				widget.hide()
				return
			if not widget.is_visible():
				widget.show()
			# Grab values
			ax, ay, aw, trash = self.background.get_area_position(area)
			cw = widget.get_allocation().width
			# Compute center
			x, y = ax + aw * 0.5 - cw * 0.5, ay + 1.0 - cw * 0.5
			# Add pad position
			x += data[0] * aw / STICK_PAD_MAX * 0.5
			y -= data[1] * aw / STICK_PAD_MAX * 0.5
			# Move circle
			self.main_area.move(widget, x, y)
		elif what in ("LT", "RT", "STICKPRESS"):
			if data[0]:
				self.hilights[App.OBSERVE_COLOR].add(what)
			else:
				self.hilights[App.OBSERVE_COLOR].remove(what)
			self._update_background()
		elif hasattr(SCButtons, what):
			try:
				if data[0]:
					self.hilights[App.OBSERVE_COLOR].add(what)
				else:
					self.hilights[App.OBSERVE_COLOR].remove(what)
				self._update_background()
			except KeyError:
				# Non fatal
				pass
		else:
			print("event", what)

	def on_profile_right_clicked(self, ps):
		for name in ("mnuConfigureController", "mnuTurnoffController"):
			# Disable controller-related menu items if controller is not connected
			obj = self.builder.get_object(name)
			obj.set_sensitive(ps.get_controller() is not None)

		for name in (
			"mnuProfileNew",
			"mnuProfileCopy",
			"mnuProfileRename",
			"mnuProfileDetails",
			"mnuProfileSeparator1",
			"mnuProfileSeparator2",
		):
			# Hide profile-related menu items for all but 1st profile switcher
			obj = self.builder.get_object(name)
			obj.set_visible(ps == self.profile_switchers[0])

		if ps == self.profile_switchers[0]:
			name = ps.get_profile_name()
			is_override = profile_is_override(name)
			is_default = profile_is_default(name)
			self.builder.get_object("mnuProfileDelete").set_visible(not is_default)
			self.builder.get_object("mnuProfileRevert").set_visible(is_override)
			self.builder.get_object("mnuProfileRename").set_visible(not is_default)
		else:
			self.builder.get_object("mnuProfileDelete").set_visible(False)
			self.builder.get_object("mnuProfileRevert").set_visible(False)

		mnuPS = self.builder.get_object("mnuPS")
		mnuPS.ps = ps
		mnuPS.popup(None, None, None, None, 3, Gtk.get_current_event_time())

	def on_mnuConfigureController_activate(self, *a):
		from scc.gui.controller_settings import ControllerSettings

		mnuPS = self.builder.get_object("mnuPS")
		cs = ControllerSettings(self, mnuPS.ps.get_controller(), mnuPS.ps)
		cs.show(self.window)

	def on_mnuProfileNew_activate(self, *a):
		mnuPS = self.builder.get_object("mnuPS")
		self.on_new_clicked(mnuPS.ps, mnuPS.ps.get_name())

	def on_mnuProfileCopy_activate(self, *a):
		mnuPS = self.builder.get_object("mnuPS")
		rbCopyProfile = self.builder.get_object("rbCopyProfile")
		self.on_new_clicked(mnuPS.ps, mnuPS.ps.get_profile_name())
		rbCopyProfile.set_active(True)

	def on_mnuProfileDetails_activate(self, *a):
		self.builder.get_object("dlgProfileDetails").show()

	def on_mnuProfileRename_activate(self, *a):
		dlg = self.builder.get_object("dlgRenameProfile")
		txRename = self.builder.get_object("txRename")
		mnuPS = self.builder.get_object("mnuPS")
		name = mnuPS.ps.get_profile_name()
		txRename.set_text(name)
		dlg._name = name
		dlg.set_transient_for(self.window)
		dlg.show()

	def on_txRename_changed(self, tx):
		name = tx.get_text()
		btRenameProfile = self.builder.get_object("btRenameProfile")
		btRenameProfile.set_sensitive(find_profile(name) is None)

	def on_btRenameProfile_clicked(self, *a):
		dlg = self.builder.get_object("dlgRenameProfile")
		txRename = self.builder.get_object("txRename")
		old_name = dlg._name
		new_name = txRename.get_text()
		old_fname = os.path.join(get_profiles_path(), old_name + ".sccprofile")
		new_fname = os.path.join(get_profiles_path(), new_name + ".sccprofile")
		try:
			os.rename(old_fname, new_fname)
			for n in (old_fname, new_fname):
				try:
					os.unlink(n + ".mod")
				except:
					# non-existing .mod file is expected
					pass
		except Exception as e:
			log.error("Failed to rename %s: %s", old_fname, e)

		controllers = list(self.dm.get_controllers())
		for c in controllers:
			if get_profile_name(c.get_profile()) == old_name:
				ps = self.profile_switchers[controllers.index(c)]
				ps.set_profile(new_name, True)
				c.set_profile(new_name)
		self.load_profile_list()
		dlg.hide()

	def on_mnuProfileDelete_activate(self, *a):
		mnuPS = self.builder.get_object("mnuPS")
		name = mnuPS.ps.get_profile_name()
		is_override = profile_is_override(name)

		if is_override:
			text = _("Really revert current profile to default values?")
		else:
			text = _("Really delete current profile?")

		d = Gtk.MessageDialog(
			parent=self.window,
			flags=Gtk.DialogFlags.MODAL,
			type=Gtk.MessageType.WARNING,
			buttons=Gtk.ButtonsType.OK_CANCEL,
			message_format=text,
		)
		d.format_secondary_text(_("This action is not undoable!"))

		if d.run() == -5:  # OK button, no idea where is this defined...
			fname = os.path.join(get_profiles_path(), name + ".sccprofile")
			try:
				os.unlink(fname)
				try:
					os.unlink(fname + ".mod")
				except:
					# non-existing .mod file is expected
					pass
				for ps in self.profile_switchers:
					ps.refresh_profile_path(name)
			except Exception as e:
				log.error("Failed to remove %s: %s", fname, e)
		d.destroy()

	def mnuTurnoffController_activate(self, *a):
		mnuPS = self.builder.get_object("mnuPS")
		if mnuPS.ps.get_controller():
			mnuPS.ps.get_controller().turnoff()

	def on_window_key_press_event(self, window, event):
		if (event.state & Gdk.ModifierType.CONTROL_MASK) != 0:
			if event.keyval == 115:
				self.on_save_clicked()
		elif self.osd_mode and event.keyval == 65471:
			self.on_save_clicked()

	def show_error(self, message, ribar=None):
		if self.ribar is None or self.ribar.get_label() is None:
			self.ribar = ribar or RIBar(message, Gtk.MessageType.ERROR)
			content = self.builder.get_object("content")
			content.pack_start(self.ribar, False, False, 0)
			content.reorder_child(self.ribar, 0)
			self.ribar.connect("close", self.hide_error)
			self.ribar.connect("response", self.hide_error)
		else:
			self.ribar.get_label().set_markup(message)
		self.ribar.show()
		self.ribar.set_reveal_child(True)
		return self.ribar

	def hide_error(self, *a):
		if self.ribar is not None:
			if self.ribar.get_parent() is not None:
				self.ribar.get_parent().remove(self.ribar)
		self.ribar = None

	def on_daemon_reconfigured(self, *a):
		log.debug("Reloading config...")
		self.config.reload()
		for ps in self.profile_switchers:
			ps.set_controller(ps.get_controller())

	def on_daemon_dead(self, *a):
		if self.just_started:
			self.dm.restart()
			self.just_started = False
			self.set_daemon_status("unknown", True)
			return

		if self.osd_mode:
			self.quit()

		for ps in self.profile_switchers:
			ps.set_controller(None)
			ps.on_daemon_dead()
		self.set_daemon_status("dead", False)

	def on_mnuEmulationEnabled_toggled(self, cb):
		if self.recursing:
			return
		if cb.get_active():
			# Turning daemon on
			self.set_daemon_status("unknown", True)
			cb.set_sensitive(False)
			self.dm.start()
		else:
			# Turning daemon off
			self.set_daemon_status("unknown", False)
			cb.set_sensitive(False)
			self.hide_error()
			self.dm.stop()

	def do_startup(self, *a) -> None:
		Gtk.Application.do_startup(self, *a)
		self.load_profile_list()
		self.setup_widgets()
		if self.app.config["gui"]["enable_status_icon"]:
			self.setup_statusicon()
		self.set_daemon_status("unknown", True)

	def do_local_options(self, trash, lo):
		set_logging_level(lo.contains("verbose"), lo.contains("debug"))
		self.osd_mode = lo.contains("osd")
		return -1

	def do_command_line(self, cl):
		Gtk.Application.do_command_line(self, cl)
		if len(cl.get_arguments()) > 1:
			filename = " ".join(cl.get_arguments()[1:])  # 'cos fuck Gtk...
			from scc.gui.importexport.dialog import Dialog

			if Dialog.determine_type(filename) is not None:
				ied = Dialog(self)

				def i_told_you_to_quit(*a):
					sys.exit(0)

				ied.window.connect("destroy", i_told_you_to_quit)
				ied.show(self.window)
				# Skip first screen and try to import this file
				ied.import_file(filename)
			else:
				sys.exit(1)
		else:
			self.activate()
		return 0

	def do_activate(self, *a):
		self.builder.get_object("window").show()
		if self.config["gui"]["minimize_on_start"] and self.statusicon and self.statusicon.get_property("active"):
			self.builder.get_object("window").hide()
		else:
			self.builder.get_object("window").show()

	def remove_dot_profile(self):
		"""Checks if first profile in list begins with dot and if yes, removes it.
		This is done to undo automatic addition that is done when daemon reports
		selecting such profile.
		"""
		cb = self.builder.get_object("cbProfile")
		model = cb.get_model()
		if len(model) == 0:
			# Nothing to remove
			return
		if not model[0][0].startswith("."):
			# Not dot profile
			return
		active = model.get_path(cb.get_active_iter())
		first = model[0].path
		if active == first:
			# Can't remove active item
			return
		model.remove(model[0].iter)

	def get_current_profile(self):
		return self.profile_switchers[0].get_profile_name()

	def set_daemon_status(self, status, daemon_runs):
		"""Updates image that shows daemon status and menu shown when image is clicked"""
		log.debug("daemon status: %s", status)
		icon = os.path.join(self.imagepath, "scc-%s.svg" % (status,))
		imgDaemonStatus = self.builder.get_object("imgDaemonStatus")
		btDaemon = self.builder.get_object("btDaemon")
		mnuEmulationEnabled = self.builder.get_object("mnuEmulationEnabled")
		mnuEmulationEnabledTray = self.builder.get_object("mnuEmulationEnabledTray")
		imgDaemonStatus.set_from_file(icon)
		mnuEmulationEnabled.set_sensitive(True)
		mnuEmulationEnabledTray.set_sensitive(True)
		self.window.set_icon_from_file(icon)
		self.status = status
		if self.statusicon:
			GLib.idle_add(self.statusicon.set, "scc-%s" % (self.status,), _("SC Controller"))
		self.recursing = True
		if status == "alive":
			btDaemon.set_tooltip_text(_("Emulation is active"))
		elif status == "error":
			btDaemon.set_tooltip_text(_("Error enabling emulation"))
		elif status == "dead":
			btDaemon.set_tooltip_text(_("Emulation is inactive"))
		else:
			btDaemon.set_tooltip_text(_("Checking emulation status..."))
		mnuEmulationEnabled.set_active(daemon_runs)
		mnuEmulationEnabledTray.set_active(daemon_runs)
		self.recursing = False

	def on_btCloseDetails_clicked(self, *a):
		self.builder.get_object("dlgProfileDetails").hide()

	def on_buffProfileDescription_changed(self, buffer, *a):
		if self.recursing:
			return
		self.current.description = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), True)
		self.on_profile_modified()

	def on_cbProfileIsTemplate_toggled(self, widget, *a):
		if self.recursing:
			return
		self.current.is_template = widget.get_active()
		self.on_profile_modified()

	def setup_commandline(self):
		def aso(long_name, short_name, description, arg=None, flags=GLib.OptionFlags.IN_MAIN):
			"""add_simple_option, adds program argument in simple way"""
			o = GLib.OptionEntry()
			if short_name:
				o.long_name = long_name
				o.short_name = short_name
			o.description = description
			o.flags = flags
			if arg is not None:
				o.arg = arg
			self.add_main_option_entries([o])

		self.connect("handle-local-options", self.do_local_options)

		aso("verbose", b"v", "Be verbose")
		aso("debug", b"d", "Be more verbose (debug mode)")
		aso("osd", b"o", "OSD mode (OSD-controllable editor for current profile)")

	def save_profile_selection(self, path):
		"""Saves name of profile into config file"""
		name = os.path.split(path)[-1]
		if name.endswith(".sccprofile"):
			name = name[0:-11]

		data = dict(current_profile=name)
		jstr = json.dumps(data, sort_keys=True, indent=4)

		open(os.path.join(get_config_path(), self.CONFIG), "w").write(jstr)

	def load_profile_selection(self):
		"""Returns name profile from config file or None if there is none saved"""
		try:
			return self.config["recent_profiles"][0]
		except Exception:
			return None

	@staticmethod
	def get_release(n: int = 4) -> str:
		"""Returns current version rounded to max. 'n' numbers.
		( v0.14.1.3 ; n=3 -> v0.14.1   )
		( v0.14.0.0 ; n=3 -> v0.14.0.0 )
		"""
		split = DAEMON_VERSION.split(".")[0:n]
		# Remove final zeroes ( v0.14.0.0 ; n=3 -> v0.14 ) - disabled, let's include them
		# while split[-1] == "0":
		# split = split[0:len(split) - 1]
		return ".".join(split)

	def release_notes_visible(self) -> bool:
		"""Returns True if release notes infobox is visible"""
		if not self.ribar:
			return False
		riNewRelease = self.builder.get_object("riNewRelease")
		return self.ribar._infobar == riNewRelease

	def check_release_notes(self):
		"""Silently downloads release notes from github and displays infobar
		informing user that they are ready to be displayed.
		"""
		url = App.RELEASE_URL % (App.get_release(),)
		log.debug(f"Loading release notes from '{url}'")
		f = Gio.File.new_for_uri(url)
		buffer = b""

		def stream_ready(stream, task, buffer):
			try:
				bytes = stream.read_bytes_finish(task)
				if bytes.get_size() > 0:
					buffer += bytes.get_data()
					stream.read_bytes_async(102400, 0, None, stream_ready, buffer)
				else:
					self.on_got_release_notes(buffer.decode("utf-8"))
			except Exception as e:
				log.warning(f"Failed to read release notes at {url}, maybe your internet connection is down?")
				log.exception(e)
				return

		def http_ready(f, task, buffer):
			try:
				stream = f.read_finish(task)
				assert stream
				stream.read_bytes_async(102400, 0, None, stream_ready, buffer)
			except Exception:
				log.warning(f"Failed to read release notes at {url}, maybe your internet connection is down?")
				# log.exception(f"Following Traceback error is not fatal and can be ignored: {e}")
				return

		f.read_async(0, None, http_ready, buffer)

	def on_got_release_notes(self, data):
		""" " Called after entire HTML page of release notes is downloaded"""
		# There is actually only one thing parsed here;
		# Sequence of words "see ... for more", in bold, containing <A> tag.
		# If such sequence is found, it's displayed with message about extended
		# release notes. Otherwise, shorter text and link to github is used.
		RE_EXTENDED = r"<strong>see.*href=\"([^\"]+).*for more.*</strong>"

		if self.ribar is not None:
			# There is already some error displayed, don't bother now...
			return

		msg = ""
		extended = re.search(RE_EXTENDED, data, re.IGNORECASE)
		if extended:
			msg += _("<a href='%s'>Click here</a> to check what's new!")
			msg = msg % (extended.group(1),)
		else:
			url = App.RELEASE_URL % (App.get_release(),)
			msg += _("Welcome to the version <b>%s</b>.")
			msg += " " + _("<a href='%s'>Click here</a> to read release notes.")
			msg = msg % (App.get_release(), url)

		infobar = self.builder.get_object("riNewRelease")
		lblNewRelease = self.builder.get_object("lblNewRelease")
		lblNewRelease.set_markup(msg)
		ribar = RIBar(None, infobar=infobar)
		ribar = self.show_error(None, ribar=ribar)
		self.ribar.connect("close", self.on_new_release_dismissed)
		self.ribar.connect("response", self.on_new_release_dismissed)

	def on_new_release_dismissed(self, *a):
		self.config["gui"]["news"]["last_version"] = App.get_release()
		self.config.save()

	def on_cbNewRelease_toggled(self, cb):
		self.app.config["gui"]["news"]["enabled"] = cb.get_active()
		self.config.save()

	def on_drag_data_received(self, widget, context, x, y, data, info, time):
		"""Drag-n-drop handler"""
		uri = None
		if str(data.get_data_type()) == "text/uri-list":
			# Only file can be dropped here
			if len(data.get_uris()):
				uri = data.get_uris()[0]
		elif str(data.get_data_type()) == "text/plain":
			# This can be anything, so try to extract uri from it
			lines = str(data.get_data()).split("\n")
			if len(lines) > 0:
				first = lines[0]
				if first.startswith("http://") or first.startswith("https://") or first.startswith("ftp://"):
					# I don't like other protocols
					uri = first
		if uri:
			from scc.gui.importexport.dialog import Dialog

			giofile = None
			if uri.startswith("file://"):
				giofile = Gio.File.new_for_uri(uri)
			else:
				# Local file can be used directly, remote has to
				# be downloaded first
				if uri.startswith("https://github.com/"):
					# Convert link to repository display to link to raw file
					uri = uri.replace("https://github.com/", "https://raw.githubusercontent.com/").replace(
						"/blob/", "/",
					)
				name = urllib.unquote(".".join(uri.split("/")[-1].split(".")[0:-1]))
				remote = Gio.File.new_for_uri(uri)
				tmp, stream = Gio.File.new_tmp("%s.XXXXXX" % (name,))
				stream.close()
				if remote.copy(tmp, Gio.FileCopyFlags.OVERWRITE, None, None):
					# Sucessfully downloaded
					log.info("Downloaded '%s'" % (uri,))
					giofile = tmp
				else:
					# Failed. Just do nothing
					return
			if giofile.get_path():
				path = giofile.get_path()
				filetype = Dialog.determine_type(path)
				if filetype:
					log.info("Importing '%s'..." % (filetype))
					log.debug("(type %s)" % (filetype,))
					ied = Dialog(self)
					ied.show(self.window)
					# Skip first screen and try to import this file
					ied.import_file(path, filetype=filetype)
				else:
					log.error("Unknown file type: '%s'..." % (path,))

	def convert_old_profiles(self):
		"""Checks all available profiles and automatically converts anything with
		version 1.3 or lower.
		"""
		from scc.parser import ActionParser

		to_convert = {}
		for name in os.listdir(get_profiles_path()):
			if name.endswith("~"):
				# Ignore backups - https://github.com/kozec/sc-controller/issues/440
				continue
			try:
				p = Profile(ActionParser())
				p.load(os.path.join(get_profiles_path(), name))
			except:
				# Just ignore invalid profiles here
				continue
			if p.original_version < 1.4:
				to_convert[name] = p

		if to_convert:
			log.warning("Auto-converting old profile files to version 1.4. This should take only moment.")
			log.warning(
				"All files are modified in-place, but backup files are created. Feel free to remove them later.",
			)
			for name in to_convert:
				try:
					to_convert[name].save("%s/%s.convert" % (get_profiles_path(), name))
					os.rename("%s/%s" % (get_profiles_path(), name), "%s/%s~" % (get_profiles_path(), name))
					os.rename("%s/%s.convert" % (get_profiles_path(), name), "%s/%s" % (get_profiles_path(), name))
					log.warning("Converted %s (from v%s)", name, to_convert[name].original_version)
				except Exception as e:
					log.warning("Failed to convert %s: %s", name, e)


class UndoRedo:
	"""Just dummy container"""

	def __init__(self, id, before, after):
		self.id = id
		self.before = before
		self.after = after



#!/usr/bin/env python3
"""SC-Controller - Horisontal OSD Menu

Works as OSD menu, but displays all items in one row.

Designed mainly as RPG numeric pad display and looks
awfull with larger number of items.
"""


import logging

from gi.repository import Gtk

from scc.constants import STICK_PAD_MIN
from scc.menu_data import Separator, Submenu
from scc.osd.grid_menu import GridMenu
from scc.osd.menu import MenuIcon

log = logging.getLogger("osd.hmenu")


class HorizontalMenu(GridMenu):
	def __init__(self, cls="osd-menu"):
		GridMenu.__init__(self, cls)

	def create_parent(self):
		g = Gtk.Grid()
		g.set_name("osd-menu")
		return g

	def generate_widget(self, item):
		"""Generates gtk widget for specified menutitem
		Ignores Submenus and Separators but applies icon size
		"""
		if isinstance(item, (Separator, Submenu)) or item.id is None:
			return None
		widget = GridMenu.generate_widget(self, item)
		icon = widget.get_children()[-1]
		if self._size > 1 and isinstance(icon, MenuIcon):
			widget.set_size_request(-1, 32 + self._size * 3)
		return widget

	def pack_items(self, parent, items):
		x = 0
		for item in items:
			parent.attach(item.widget, x, 0, 1, 1)
			x += 1

	def on_stick_direction(self, trash, x, y):
		if x != 0:
			self.next_item(-x)

	def on_event(self, daemon, what, data):
		# Restricts Y axis to dead center, as nothing
		# else makes sense in this kind of menu
		if self._submenu:
			return self._submenu.on_event(daemon, what, data)
		if what == self._control_with and self._use_cursor:
			data = data[0], STICK_PAD_MIN
		GridMenu.on_event(self, daemon, what, data)



"""Common code for all (one) USB-based drivers.

Driver that uses USB has to call
register_hotplug_device(callback, vendor_id, product_id, on_failure=None)
method to get notified about connected USB devices.

Callback will be called with following arguments:
	callback(device, handle)
Callback has to return created USBDevice instance or None.
"""

from __future__ import annotations

import logging
import time
import traceback
from typing import TYPE_CHECKING

import usb1

if TYPE_CHECKING:
	from usb1 import USBDeviceHandle, USBTransfer

	from scc.sccdaemon import SCCDaemon

log = logging.getLogger("USB")


class USBDevice:
	"""Base class for all handled usb devices."""

	def __init__(self, device: USBDevice, handle: USBDeviceHandle) -> None:
		self.device = device
		self.handle = handle
		self._claimed = []
		self._cmsg = []  # controll messages
		self._rmsg = []  # requests (excepts response)
		self._transfer_list = []

	def set_input_interrupt(self, endpoint: int, size: int, callback) -> None:
		"""Set up input transfer

		callback(endpoint, data) is called repeadedly with every packed received.
		"""

		def callback_wrapper(transfer: USBTransfer) -> None:
			if transfer.getStatus() != usb1.TRANSFER_COMPLETED or transfer.getActualLength() != size:
				return

			data = transfer.getBuffer()
			try:
				callback(endpoint, data)
			except Exception as e:
				log.error("Failed to handle received data")
				log.error(e)
				log.error(traceback.format_exc())
			finally:
				try:  # https://github.com/C0rn3j/sc-controller/issues/57
					transfer.submit()
				except Exception:
					log.exception("Failed to submit the transfer!")

		transfer = self.handle.getTransfer()
		transfer.setInterrupt(
			usb1.ENDPOINT_IN | endpoint,
			size,
			callback=callback_wrapper,
		)
		transfer.submit()
		self._transfer_list.append(transfer)

	def send_control(self, index, data):
		"""Schedules writing control to device."""
		zeros = b"\x00" * (64 - len(data))

		self._cmsg.insert(
			0,
			(
				0x21,  # request_type
				0x09,  # request
				0x0300,  # value
				index,
				data + zeros,
				0,  # Timeout
			),
		)

	def overwrite_control(self, index, data):
		"""Similar to send_control, but this one checks and overwrites already scheduled controll for same device/index."""
		for x in self._cmsg:
			x_index, x_data, x_timeout = x[-3:]
			# First 3 bytes are for PacketType, size and ConfigType
			if x_index == index and x_data[0:3] == data[0:3]:
				self._cmsg.remove(x)
				break
		self.send_control(index, data)

	def make_request(self, index, callback, data, size=64):
		"""Schedule a synchronous request that requires response.

		Request is done ASAP and provided callback is called with received data.
		"""
		self._rmsg.append(
			(
				(
					0x21,  # request_type
					0x09,  # request
					0x0300,  # value
					index,
					data,
				),
				index,
				size,
				callback,
			),
		)

	def flush(self):
		"""Flush all prepared control messages to the device."""
		while len(self._cmsg):
			msg = self._cmsg.pop()
			self.handle.controlWrite(*msg)

		while len(self._rmsg):
			msg, index, size, callback = self._rmsg.pop()
			self.handle.controlWrite(*msg)
			data = self.handle.controlRead(
				0xA1,  # request_type
				0x01,  # request
				0x0300,  # value
				index,
				size,
			)
			callback(data)

	def force_restart(self):
		"""Restart device, close handle and try to re-grab it again.

		Don't use unless absolutely necessary.
		"""
		tp = self.device.getVendorID(), self.device.getProductID()
		self.close()
		_usb._retry_devices.append(tp)

	def claim(self, number: int):
		"""Remember list of claimed interfaces and allow to unclaim them all at once using unclaim() method

		or automatically when device is closed.
		"""
		self.handle.claimInterface(number)
		self._claimed.append(number)

	def claim_by(self, klass, subclass, protocol) -> int:
		"""Claim all interfaces with specified parameters.

		Return number of claimed interfaces
		"""
		rv = 0
		for inter in self.device[0]:
			for setting in inter:
				number = setting.getNumber()
				if self.handle.kernelDriverActive(number):
					self.handle.detachKernelDriver(number)
				ksp = setting.getClass(), setting.getSubClass(), setting.getProtocol()
				if ksp == (klass, subclass, protocol):
					self.claim(number)
					rv += 1
		return rv

	def unclaim(self):
		"""Unclaim all claimed interfaces."""
		for number in self._claimed:
			try:
				self.handle.releaseInterface(number)
				self.handle.attachKernelDriver(number)
			except usb1.USBErrorNoDevice:
				# Safe to ignore, happens when USB is removed
				pass
		self._claimed = []

	def close(self):
		"""Called after device is disconnected."""
		try:
			self.unclaim()
		except Exception:
			pass
		try:
			self.handle.resetDevice()
			self.handle.close()
		except Exception:
			pass


class USBDriver:
	def __init__(self):
		self.daemon = None
		self._known_ids = {}
		self._fail_cbs = {}
		self._devices = {}
		self._syspaths = {}
		self._started = False
		self._retry_devices = []
		self._retry_devices_timer = 0
		self._ctx = None  # Set by start method
		self._changed = 0

	def set_daemon(self, daemon: SCCDaemon) -> None:
		self.daemon = daemon

	def on_exit(self, *a):
		"""Close all devices and unclaim all interfaces."""
		if len(self._devices):
			log.debug("Releasing devices...")
			to_release, self._devices, self._syspaths = self._devices.values(), {}, {}
			for d in to_release:
				d.close()

	def start(self):
		self._ctx = usb1.USBContext()

		def fd_cb(*a):
			self._changed += 1

		def register_fd(fd, events, *a):
			self.daemon.get_poller().register(fd, events, fd_cb)

		def unregister_fd(fd, *a):
			self.daemon.get_poller().unregister(fd)

		self._ctx.setPollFDNotifiers(register_fd, unregister_fd)
		for fd, events in self._ctx.getPollFDList():
			register_fd(fd, events)
		self._started = True

	def handle_new_device(self, syspath: str, vendor: int, product: int) -> bool | None:
		tp = vendor, product
		handle = None
		if tp not in self._known_ids:
			return None
		bus, dev = self.daemon.get_device_monitor().get_usb_address(syspath)
		for device in self._ctx.getDeviceIterator():
			if (bus, dev) == (device.getBusNumber(), device.getDeviceAddress()):
				try:
					handle = device.open()
					break
				except usb1.USBError as e:
					log.error("Failed to open USB device %.4x:%.4x : %s", tp[0], tp[1], e)
					if tp in self._fail_cbs:
						self._fail_cbs[tp](syspath, *tp)
						return None
					if self.daemon:
						self.daemon.add_error(
							f"usb:{tp[0]}:{tp[1]}",
							f"Failed to open USB device: {e}",
						)
					return None
		else:
			return None

		callback = self._known_ids[tp]
		handled_device = None
		try:
			handled_device = callback(device, handle)
		except usb1.USBErrorBusy as e:
			log.error("Failed to claim USB device %.4x:%.4x : %s", tp[0], tp[1], e)
			if tp in self._fail_cbs:
				device.close()
				self._fail_cbs[tp](syspath, *tp)
				return False
			if self.daemon:
				self.daemon.add_error(
					f"usb:{tp[0]}:{tp[1]}",
					f"Failed to claim USB device: {e}",
				)
			self._retry_devices.append((syspath, tp))
			device.close()
			return True
		if handled_device:
			self._devices[device] = handled_device
			self._syspaths[syspath] = device
			log.debug("USB device added: %.4x:%.4x", *tp)
			self.daemon.remove_error(f"usb:{tp[0]}:{tp[1]}")
			return True
		log.warning("Known USB device ignored: %.4x:%.4x", *tp)
		device.close()
		return False

	def handle_removed_device(self, syspath: str, vendor: int, product: int) -> None:
		if syspath in self._syspaths:
			device = self._syspaths[syspath]
			handled_device = self._devices[device]
			del self._syspaths[syspath]
			del self._devices[device]
			handled_device.close()
			try:
				device.close()
			except usb1.USBErrorNoDevice:
				# Safe to ignore, happens when device is physically removed
				pass

	def register_hotplug_device(self, callback, vendor_id: int, product_id: int, on_failure) -> None:
		self._known_ids[vendor_id, product_id] = callback
		if on_failure:
			self._fail_cbs[vendor_id, product_id] = on_failure
		monitor = self.daemon.get_device_monitor()
		monitor.add_callback("usb", vendor_id, product_id, self.handle_new_device, self.handle_removed_device)
		log.debug("Registered USB driver for %.4x:%.4x", vendor_id, product_id)

	def unregister_hotplug_device(self, callback, vendor_id: int, product_id: int) -> None:
		if self._known_ids.get((vendor_id, product_id)) == callback:
			del self._known_ids[vendor_id, product_id]
			if (vendor_id, product_id) in self._fail_cbs:
				del self._fail_cbs[vendor_id, product_id]
			log.debug("Unregistred USB driver for %.4x:%.4x", vendor_id, product_id)

	def mainloop(self):
		if self._changed > 0:
			self._ctx.handleEventsTimeout()
			self._changed = 0

		for d in self._devices.values():  # TODO: don't use .values() here
			try:
				d.flush()
			except usb1.USBErrorPipe:
				log.error("USB device %s disconnected durring flush", d)
				d.close()
				break
		if len(self._retry_devices):
			if time.time() > self._retry_devices_timer:
				self._retry_devices_timer = time.time() + 5.0
				lst, self._retry_devices = self._retry_devices, []
				for syspath, (vendor, product) in lst:
					self.handle_new_device(syspath, vendor, product)


# USBDriver should be process-wide singleton
_usb = USBDriver()


def init(daemon: SCCDaemon, config: dict) -> bool:
	_usb.set_daemon(daemon)
	daemon.add_on_exit(_usb.on_exit)
	daemon.add_mainloop(_usb.mainloop)
	return True


def start(daemon: SCCDaemon) -> None:
	_usb.start()


def register_hotplug_device(callback, vendor_id: int, product_id: int, on_failure=None) -> None:
	_usb.register_hotplug_device(callback, vendor_id, product_id, on_failure)


def unregister_hotplug_device(callback, vendor_id: int, product_id: int) -> None:
	_usb.unregister_hotplug_device(callback, vendor_id, product_id)



"""SC Controller - Profile Manager.

Simple class that manages stuff related to creating, loading, listing (...) of
user-editable data - that are profiles, menus and controller-icons.

Main App class interits from this.
"""

import logging
import os

from gi.repository import Gio

from scc.gui.parser import GuiActionParser
from scc.paths import (
	get_default_menuicons_path,
	get_default_menus_path,
	get_default_profiles_path,
	get_menuicons_path,
	get_menus_path,
	get_profiles_path,
)
from scc.profile import Profile

log = logging.getLogger("UDataManager")


class UserDataManager:
	def __init__(self) -> None:
		profiles_path = get_profiles_path()
		if not os.path.exists(profiles_path):
			log.info("Creating profile directory '%s'", profiles_path)
			os.makedirs(profiles_path)

		menus_path = get_menus_path()
		if not os.path.exists(menus_path):
			log.info("Creating menu directory '%s'", menus_path)
			os.makedirs(menus_path)

		menuicons_path = get_menuicons_path()
		if not os.path.exists(menuicons_path):
			log.info("Creating menuicons directory '%s'", menuicons_path)
			os.makedirs(menuicons_path)

	def load_profile(self, giofile: Gio.File) -> None:
		"""Load profile from 'giofile' into 'profile' object.

		Calls on_profiles_loaded when done
		"""
		# This may get asynchronous later, but that load runs under 1ms...
		profile = Profile(GuiActionParser())
		profile.load(giofile.get_path())
		self.on_profile_loaded(profile, giofile)

	def save_profile(self, giofile: Gio.File, profile: Profile) -> None:
		"""Save profile from 'profile' object into 'giofile'.

		Calls on_profile_saved when done
		"""
		# 1st check, if file is not in /usr/share.
		# When user tries to save over built-in profile in /usr/share,
		# new file with same name is created in ~/.config/scc/profiles and profile
		# is shaved into it.

		if giofile.get_path().startswith(get_default_profiles_path()):
			return self._save_profile_local(giofile, profile)

		profile.save(giofile.get_path())
		self.on_profile_saved(giofile)

	def _save_profile_local(self, giofile: Gio.File, profile: Profile) -> None:
		filename = os.path.split(giofile.get_path())[-1]
		localpath = os.path.join(get_profiles_path(), filename)
		giofile = Gio.File.new_for_path(localpath)
		self.save_profile(giofile, profile)

	def load_profile_list(self, category=None) -> None:
		paths = [get_default_profiles_path(), get_profiles_path()]
		self.load_user_data(paths, "*.sccprofile", category, self.on_profiles_loaded)

	def load_menu_list(self, category=None) -> None:
		paths = [get_default_menus_path(), get_menus_path()]
		self.load_user_data(paths, "*.menu", category, self.on_menus_loaded)

	def load_menu_icons(self, category=None) -> None:
		paths = [get_default_menuicons_path(), get_menuicons_path()]
		self.load_user_data(paths, "*.png", category, self.on_menuicons_loaded)

	def load_user_data(self, paths, pattern, category, callback) -> None:
		"""Load data such as of profiles. Uses GLib to do it in the background."""
		if category:
			paths = [os.path.join(p, category) for p in paths]

		# First list is for default stuff, then for user-defined
		# Number is increased when list is loaded until it reaches 2
		data = [None] * len(paths)

		for i in range(len(paths)):
			f = Gio.File.new_for_path(paths[i])
			f.enumerate_children_async(
				pattern,
				Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
				1,
				None,
				self._on_user_data_loaded,
				data,
				i,
				callback,
			)

	def _on_user_data_loaded(self, pdir, res, data, i, callback) -> None:
		"""Called when enumerate_children_async gets lists of files.

		Usually called twice for default (system) and user directory.
		"""
		try:
			data[i] = pdir, pdir.enumerate_children_finish(res)
		except Exception as e:
			# Usually when directory doesn't exists
			log.warning("enumerate_children_finish for %s failed: %s", pdir.get_path(), e)
			data[i] = None, []
		if None not in data:
			files = {}
			try:
				for pdir, enumerator in data:
					if pdir is not None:
						for finfo in enumerator:
							name = finfo.get_name()
							if name and not name.endswith("~"):
								files[name] = pdir.get_child(name)
			except Exception as e:
				# https://github.com/kozec/sc-controller/issues/50
				log.warning("enumerate_children_async failed: %s", e)
				files = self._sync_load([pdir for pdir, enumerator in data if pdir is not None])
			if len(files) < 1:
				# https://github.com/kozec/sc-controller/issues/327
				log.warning("enumerate_children_async returned no files")
				files = self._sync_load([pdir for pdir, enumerator in data if pdir is not None])

			callback(files.values())

	def _sync_load(self, pdirs) -> dict:
		"""Synchronous (= UI lagging) fallback method for those (hopefully) rare cases when enumerate_children_finish returns nonsense."""
		files = {}
		for pdir in pdirs:
			for name in os.listdir(pdir.get_path()):
				files[name] = pdir.get_child(name)
		return files

	def on_menus_loaded(self, menus) -> None:  # Overriden by subclass
		pass

	def on_profiles_loaded(self, profiles) -> None:  # Overriden by subclass
		pass

	def on_menuicons_loaded(self, icons) -> None:  # Overriden by subclass
		pass

	def on_profile_saved(self, giofile: Gio.File) -> None:  # Overriden in App
		pass

	def on_profile_loaded(self, profile: Profile, giofile: Gio.File) -> None:  # Overriden in App
		pass



"""SC Controller - Steam Controller Driver

Driver for Steam Controller over bluetooth (evdev)

Shares a lot of classes with sc_dongle.py
"""

import ctypes
import logging
import os
import struct
import sys
from math import cos, sin

from scc.constants import STICK_PAD_MAX, STICK_PAD_MIN, ControllerFlags
from scc.lib.hidraw import HIDRaw
from scc.tools import find_library

from .sc_dongle import SCConfigType, SCController, SCPacketLength, SCPacketType

VENDOR_ID = 0x28DE
PRODUCT_ID = 0x1106
PACKET_SIZE = 20

log = logging.getLogger("SCBT")


class SCByBtControllerInput(ctypes.Structure):
	_fields_ = [
		("type", ctypes.c_uint16),
		("buttons", ctypes.c_uint32),
		("ltrig", ctypes.c_uint8),
		("rtrig", ctypes.c_uint8),
		("stick_x", ctypes.c_int32),
		("stick_y", ctypes.c_int32),
		("lpad_x", ctypes.c_int32),
		("lpad_y", ctypes.c_int32),
		("rpad_x", ctypes.c_int32),
		("rpad_y", ctypes.c_int32),
		("accel_x", ctypes.c_int32),
		("accel_y", ctypes.c_int32),
		("accel_z", ctypes.c_int32),
		("gpitch", ctypes.c_int32),
		("groll", ctypes.c_int32),
		("gyaw", ctypes.c_int32),
		("q1", ctypes.c_int32),
		("q2", ctypes.c_int32),
		("q3", ctypes.c_int32),
		("q4", ctypes.c_int32),
	]


class SCByBtC(ctypes.Structure):
	_fields_ = [
		("fileno", ctypes.c_int),
		("buffer", ctypes.c_char * 256),
		("long_packet", ctypes.c_uint8),
		("state", SCByBtControllerInput),
		("old_state", SCByBtControllerInput),
	]


SCByBtCPtr = ctypes.POINTER(SCByBtC)


class Driver:
	"""Similar to USB driver, but with hidraw used for backend"""

	# TODO: It should be possible to merge this, usb and hiddrv

	def __init__(self, daemon, config):
		self.config = config
		self.daemon = daemon
		self.reconnecting = set()
		self._lib = find_library("libsc_by_bt")
		read_input = self._lib.read_input
		read_input.restype = ctypes.c_int
		read_input.argtypes = [SCByBtCPtr]
		daemon.get_device_monitor().add_callback("bluetooth", VENDOR_ID, PRODUCT_ID, self.new_device_callback, None)

	def retry(self, syspath):
		"""Schedules reconnecting controller after read operation fails.
		"""

		def reconnect(*a):
			if syspath in self.reconnecting:
				self.reconnecting.remove(syspath)
				log.debug("Reconnecting to controller...")
				self.new_device_callback(syspath)

		self.reconnecting.add(syspath)
		self.daemon.get_device_monitor().add_remove_callback(syspath, self._retry_cancel)
		self.daemon.get_scheduler().schedule(1.0, reconnect)

	def _retry_cancel(self, syspath, vendor, product):
		"""Cancels reconnection scheduled by 'retry'. Called when device monitor
		reports controller (as in BT device) being disconencted.
		"""
		if syspath in self.reconnecting:
			self.reconnecting.remove(syspath)

	def new_device_callback(self, syspath, *whatever):
		hidrawname = self.daemon.get_device_monitor().get_hidraw(syspath)
		if hidrawname is None:
			return None
		try:
			dev = HIDRaw(open(os.path.join("/dev/", hidrawname), "w+b"))
			return SCByBt(self, syspath, dev)
		except Exception as e:
			log.exception(e)
			return None


class SCByBt(SCController):
	flags = 0 | ControllerFlags.SEPARATE_STICK

	def __init__(self, driver, syspath, hidrawdev):
		self._cmsg = []  # controll messages
		self._transfer_list = []
		self.driver = driver
		self.daemon = driver.daemon
		self.syspath = syspath
		SCController.__init__(self, self, -1, -1)
		self._led_level = 30
		self._device_name = hidrawdev.getName()
		self._hidrawdev = hidrawdev
		self._fileno = hidrawdev._device.fileno()
		self._c_data = SCByBtC(fileno=self._fileno, long_packet=0)
		self._c_data_ptr = ctypes.byref(self._c_data)
		self._old_state = self._c_data.old_state
		self._state = self._c_data.state
		self._poller = self.daemon.get_poller()
		if self._poller:
			self._poller.register(self._fileno, self._poller.POLLIN, self._input)
		self.daemon.get_device_monitor().add_remove_callback(syspath, self.close)
		self.read_serial()
		self.configure()
		self.flush()
		self.daemon.add_controller(self)

	def get_device_name(self):
		# Method needed by evdev driver
		# return self._device_name
		return "Steam Controller over Bluetooth"

	def get_type(self):
		return "scbt"

	def __repr__(self):
		return "<SCByBt %s>" % (self.get_id(),)

	def configure(self, idle_timeout=None, enable_gyros=None, led_level=None):
		"""Sets and, if possible, sends configuration to controller.
		See SCController.configure method in sc_dongle.py;

		This method is almost the same, with different set of hardcoded constants.
		"""
		# ------
		"""
		packet format:
		 - uint8_t type - SCPacketType.CONFIGURE
		 - uint8_t size - SCPacketLength.CONFIGURE_BT or SCPacketLength.LED
		 - uint8_t config_type - SCConfigType.CONFIGURE_BT or SCConfigType.LED
		 - (variable) data

		Format for data when configuring controller:
		 - 12B		unknown1 - (hex 0000310200080700070700300)
		 - uint8	enable gyro sensor - 0x14 enables, 0x00 disables
		 - 2b		unknown2 - (0x00, 0x2e)

		Format for data when configuring led:
		 - uint8	led
		 - 60b		unused
		"""
		# idle_timeout is ignored
		if enable_gyros is not None:
			self._enable_gyros = enable_gyros
		if led_level is not None:
			self._led_level = int(led_level)

		unknown1 = b"\x00\x00\x31\x02\x00\x08\x07\x00\x07\x07\x00\x30"
		unknown2 = b"\x00\x2e"

		# Timeout & Gyros
		self.overwrite_control(
			self._ccidx,
			struct.pack(
				">BBB12sB2s",
				SCPacketType.CONFIGURE,
				SCPacketLength.CONFIGURE_BT,
				SCConfigType.CONFIGURE_BT,
				unknown1,
				# 0x10 (Gyro) | 0x08 (Accel) | 0x04 (Quat)
				0x1C if self._enable_gyros else 0,
				unknown2,
			),
		)

		# LED
		self.overwrite_control(
			self._ccidx,
			struct.pack(">BBBB", SCPacketType.CONFIGURE, SCPacketLength.LED, SCConfigType.LED, self._led_level),
		)

	def read_serial(self):
		self._serial = self._hidrawdev.getPhysicalAddress().replace(b":", b"")

	def send_control(self, index, data):
		"""Schedules writing control to device"""
		# For BT controller, index is ignored
		zeros = b"\x00" * (PACKET_SIZE - len(data) - 1)
		self._cmsg.insert(0, b"\xc0" + data + zeros)

	def overwrite_control(self, index, data):
		"""Similar to send_control, but this one checks and overwrites
		already scheduled controll for same device/index.
		"""
		# For BT controller, index is ignored
		for x in self._cmsg:
			# First byte is reserved, following 3 are for PacketType, size and ConfigType
			if x[0:4] == data[0:4]:
				self._cmsg.remove(x)
				break
		self.send_control(index, data)

	def make_request(self, index, callback, data, size=PACKET_SIZE):
		"""There are no requests one can send to BT controller,
		so this just causes exception.
		"""
		raise RuntimeError("make_request over BT not implemented")

	def flush(self):
		"""Flushes all prepared control messages to the device"""
		while len(self._cmsg):
			msg = self._cmsg.pop()
			# Feature report data must be sent with report ID 3
			# or Input/output error will occur with later BlueZ versions (5.64)
			# Does not affect older BlueZ versions
			self._hidrawdev.sendFeatureReport(msg, 3)

	def input(self, idata):
		raise RuntimeError("This shouldn't be called, ever")

	def turnoff(self):
		super().turnoff()
		# Need to call flush to make sure packet
		# is sent to controller
		self.flush()

	def close(self, *a):
		if self._poller:
			self._poller.unregister(self._fileno)
		self.daemon.remove_controller(self)
		self._hidrawdev._device.close()

	def disconnected(self):
		pass

	def _input(self, *a):
		r = self.driver._lib.read_input(self._c_data_ptr)

		if r == 1:
			if self.mapper is not None:
				if self._input_rotation_l and (self._state.type & 0x0100) != 0:
					lx, ly = self._state.lpad_x, self._state.lpad_y
					s, c = sin(self._input_rotation_l), cos(self._input_rotation_l)

					# Adjust LX for rotation and clamp
					value = int(lx * c - ly * s)
					self._state.lpad_x = max(STICK_PAD_MIN, min(STICK_PAD_MAX, value))

					# Adjust LY for rotation and clamp
					value = int(lx * s + ly * c)
					self._state.lpad_y = max(STICK_PAD_MIN, min(STICK_PAD_MAX, value))
				if self._input_rotation_r and (self._state.type & 0x0200) != 0:
					rx, ry = self._state.rpad_x, self._state.rpad_y
					s, c = sin(self._input_rotation_r), cos(self._input_rotation_r)

					# Adjust RX for rotation and clamp
					value = int(rx * c - ry * s)
					self._state.rpad_x = max(STICK_PAD_MIN, min(STICK_PAD_MAX, value))

					# Adjust RY for rotation and clamp
					value = int(rx * s + ry * c)
					self._state.rpad_y = max(STICK_PAD_MIN, min(STICK_PAD_MAX, value))

				self.mapper.input(self, self._old_state, self._state)
			self.flush()
		elif r > 1:
			log.error("Read Failed")
			self.close()
			self.driver.retry(self.syspath)


def hidraw_test(filename):
	class FakeDaemon:
		def add_error(self, id, error):
			log.error(error)

		def remove_error(*a):
			pass

		def add_mainloop(*a):
			pass

		def get_active_ids(*a):
			return []

		def get_poller(self):
			return None

	class TestSC(SCByBt):
		def input(self, tup):
			print(tup)

	dev = HIDRaw(open(filename, "w+b"))
	driver = Driver(FakeDaemon(), {})
	c = TestSC(driver, None, dev)
	c.configure()
	c.flush()
	while True:
		c._input()
		print({x[0]: getattr(c._state, x[0]) for x in c._state._fields_})


_drv = None


def init(daemon, config):
	"""Registers hotplug callback for controller dongle"""
	# if not (HAVE_EVDEV and config["drivers"].get("evdevdrv")):
	# 	log.warning("Evdev driver is not enabled, Steam Controller over Bluetooth support cannot be enabled.")
	# 	return False
	_drv = Driver(daemon, config)
	return True


if __name__ == "__main__":
	""" Called when executed as script """
	from scc.tools import init_logging, set_logging_level

	init_logging()
	set_logging_level(True, True)
	sys.exit(hidraw_test(sys.argv[1]))



#!/usr/bin/env python3
"""SC-Controller - Gestures

Everything related to non-GUI part of gesture detection lies here.
It's technically part of SCC-Daemon, separater into special module just to keep
it clean.
"""

import logging

from scc.actions import Action
from scc.constants import CPAD, CPAD_MIN, CPAD_X_MAX, CPAD_Y_MAX, STICK_PAD_MAX, STICK_PAD_MIN
from scc.tools import clamp

log = logging.getLogger("Gestures")


class GestureDetector(Action):
	"""Derived from Action, but not callable in profile.

	When daemon decides it's good time to start gesture, be it because of
	GestureAction special action or "Gesture:" message from client,
	it constructs instance of this class and leaves everything to it.
	"""

	UP = "U"
	DOWN = "D"
	LEFT = "L"
	RIGHT = "R"

	def __init__(self, up_direction, on_finished):
		Action.__init__(self)
		# TODO: Configurable resolution
		self._resolution = 3
		self._deadzone = 1.0 / self._resolution / self._resolution
		self._up_direction = up_direction
		self._on_finished = on_finished
		self._enabled = False
		self._positions = []
		self._result = []

	def enable(self):
		"""GestureDetector doesn't starts do detect anything until this is called"""
		self._enabled = True
		self._result = []

	def get_string(self):
		"""Returns string representation of (probably unfinished) gesture"""
		return "".join(self._result)

	def get_positions(self):
		"""Returns list of positions used to generate gesture"""
		return self._positions

	def get_resolution(self):
		"""Returns gesture resolution"""
		return self._resolution

	def whole(self, mapper, x, y, what):
		if self._enabled:
			if (x, y) == (0, 0):
				# Pad was released
				self._enabled = False
				self._on_finished(self, "".join(self._result))
				return
			# Convert positions on pad to position on grid
			if what == CPAD:
				x = clamp(0, float(x) / (CPAD_X_MAX - CPAD_MIN), 1.0)
				y = clamp(0, float(y) / (CPAD_Y_MAX - CPAD_MIN), 1.0)
				x *= self._resolution
				y *= self._resolution
			else:
				x -= STICK_PAD_MIN
				y = STICK_PAD_MAX - y
				x = float(x) / (float(STICK_PAD_MAX - STICK_PAD_MIN) / self._resolution)
				y = float(y) / (float(STICK_PAD_MAX - STICK_PAD_MIN) / self._resolution)
			# Check for deadzones around grid lines
			for i in range(1, self._resolution):
				if x > i - self._deadzone and x < i + self._deadzone:
					return
				if y > i - self._deadzone and y < i + self._deadzone:
					return
			# Round
			x = clamp(0, int(x), self._resolution - 1)
			y = clamp(0, int(y), self._resolution - 1)
			if self._positions:
				ox, oy = self._positions[-1]
				if (x, y) != (ox, oy):
					self._positions.append((x, y))
					while (x, y) != (ox, oy):
						if x < ox:
							self._result.append(self.LEFT)
							x += 1
						elif x > ox:
							self._result.append(self.RIGHT)
							x -= 1
						elif y < oy:
							self._result.append(self.UP)
							y += 1
						elif y > oy:
							self._result.append(self.DOWN)
							y -= 1
			else:
				self._positions.append((x, y))



#!/usr/bin/env python3
"""SC-Controller - Action Editor

Allows to edit button or trigger action.
"""

import logging

from gi.repository import Gtk

from scc.actions import Action, NoAction, RangeOP
from scc.constants import HapticPos, SCButtons
from scc.gui.controller_widget import PADS, STICKS
from scc.gui.dwsnc import headerbar
from scc.gui.editor import Editor
from scc.macros import Macro
from scc.modifiers import DoubleclickModifier, FeedbackModifier, HoldModifier, ModeModifier
from scc.tools import _, nameof

log = logging.getLogger("ModeshiftEditor")


class ModeshiftEditor(Editor):
	GLADE = "modeshift_editor.glade"
	BUTTONS = (  # in order as displayed in combobox
		(SCButtons.A, _("A")),
		(SCButtons.B, _("B")),
		(SCButtons.X, _("X")),
		(SCButtons.Y, _("Y")),
		(None, None),
		(SCButtons.BACK, _("Back (select)")),
		(SCButtons.C, _("Center")),
		(SCButtons.START, _("Start")),
		(None, None),
		(SCButtons.LGRIP, _("Left Grip")),
		(SCButtons.RGRIP, _("Right Grip")),
		(SCButtons.LB, _("Left Bumper")),
		(SCButtons.RB, _("Right Bumper")),
		(None, None),
		(SCButtons.LT, _("Left Trigger (full)")),
		(SCButtons.RT, _("Right Trigger (full)")),
		("Soft LT", _("Left Trigger (soft)")),
		("Soft RT", _("Right Trigger (soft)")),
		(None, None),
		(SCButtons.STICKPRESS, _("Stick Pressed")),
		(SCButtons.LPAD, _("Left Pad Pressed")),
		(SCButtons.RPAD, _("Right Pad Pressed")),
		(SCButtons.LPADTOUCH, _("Left Pad Touched")),
		(SCButtons.RPADTOUCH, _("Right Pad Touched")),
	)

	def __init__(self, app, callback):
		Editor.__init__(self)
		self.app = app
		self.id = None
		self.mode = Action.AC_BUTTON
		self.ac_callback = callback
		self.soft_level = -1
		self.current_page = 0
		self.actions = ([], [], [])
		self.nomods = [NoAction(), NoAction(), NoAction()]
		self.setup_widgets()

	def setup_widgets(self):
		Editor.setup_widgets(self)

		cbButtonChooser = self.builder.get_object("cbButtonChooser")
		cbButtonChooser.set_row_separator_func(lambda model, iter: model.get_value(iter, 0) is None)

		b = lambda a: self.builder.get_object(a)
		self.action_widgets = (
			# Order goes: Grid, 1st Action Button, Clear Button
			# 1st group, 'pressed'
			(b("grActions"), b("btDefault"), b("btClearDefault")),
			# 2nd group, 'hold'
			(b("grHold"), b("btHold"), b("btClearHold")),
			# 2nd group, 'double-click'
			(b("grDoubleClick"), b("btDoubleClick"), b("btClearDoubleClick")),
		)

		headerbar(self.builder.get_object("header"))

	def on_Dialog_destroy(self, *a):
		self.remove_added_widget()

	def _fill_button_chooser(self, *a):
		cbButtonChooser = self.builder.get_object("cbButtonChooser")
		model = cbButtonChooser.get_model()
		model.clear()
		for item, text in self.BUTTONS:
			if any([x[0] == item for x in self.actions[self.current_page]]):
				# Skip already added buttons
				continue
			if type(item) is str:
				# Special case for soft pull items
				button = getattr(SCButtons, item.split(" ")[-1])
				if any([(isinstance(x, RangeOP) and x.what == button) for x in self.actions[self.current_page]]):
					# Skip already added soft pulls
					continue

			if item == SCButtons.STICKPRESS:
				if self.id == nameof(SCButtons.LPAD):
					# Controller cannot handle pressing stick and lpad at once
					continue
			model.append((None if item is None else nameof(item), text))
		cbButtonChooser.set_active(0)

	def _add_action(self, index, what, action):
		cbButtonChooser = self.builder.get_object("cbButtonChooser")
		adjSoftLevel = self.builder.get_object("adjSoftLevel")
		rvSoftLevel = self.builder.get_object("rvSoftLevel")
		grActions = self.action_widgets[index][0]
		model = cbButtonChooser.get_model()

		for row in model:
			item = model.get_value(row.iter, 0)
			if item:
				if item == nameof(what):
					model.remove(row.iter)
					break
				if isinstance(what, RangeOP) and item.startswith("Soft"):
					if nameof(what.what) == item.split(" ")[-1]:
						model.remove(row.iter)
						break
		try:
			while model.get_value(model[0].iter, 0) is None:
				model.remove(model[0].iter)
			cbButtonChooser.set_active(0)
		except:
			pass

		i = len(self.actions[index]) + 1
		l = Gtk.Label()
		if isinstance(what, RangeOP) and what.op == ">=":
			if self.soft_level == -1:
				# First Range added, load level from it
				self.soft_level = what.value if what.value > 0.0 else 0.75
				adjSoftLevel.set_value(self.soft_level)
				rvSoftLevel.set_reveal_child(True)
				what.value = -1
			if what.value == -1:
				# Range with value taken from "Soft Pull Level" slider
				l.set_markup("<b>%s (soft pull)</b>" % (nameof(what.what),))
			else:
				# Any other range
				l.set_markup("<b>%s</b>" % (nameof(what),))
		else:
			l.set_markup("<b>%s</b>" % (nameof(what),))
		l.set_xalign(0.0)
		b = Gtk.Button.new_with_label(action.describe(self.mode))
		b.set_property("hexpand", True)
		b.connect("clicked", self.on_actionb_clicked, index, what)
		clearb = Gtk.Button()
		clearb.set_image(Gtk.Image.new_from_stock("gtk-delete", Gtk.IconSize.SMALL_TOOLBAR))
		clearb.set_relief(Gtk.ReliefStyle.NONE)
		clearb.connect("clicked", self.on_clearb_clicked, index, what)
		grActions.attach(l, 0, i, 1, 1)
		grActions.attach(b, 1, i, 1, 1)
		grActions.attach(clearb, 2, i, 1, 1)

		self.actions[index].append([what, action, l, b, clearb])
		grActions.show_all()

	def on_clearb_clicked(self, trash, index, button):
		grActions = self.action_widgets[index][0]
		cbButtonChooser = self.builder.get_object("cbButtonChooser")
		model = cbButtonChooser.get_model()
		# Remove requested action from the list
		for i in range(len(self.actions[index])):
			if self.actions[index][i][0] == button:
				button, action, l, b, clearb = self.actions[index][i]
				for w in (l, b, clearb):
					grActions.remove(w)
				del self.actions[index][i]
				break
		# Move everything after that action one position up
		# - remove it
		for j in range(i, len(self.actions[index])):
			button, action, l, b, clearb = self.actions[index][j]
			for w in (l, b, clearb):
				grActions.remove(w)
		# - add it again
		for j in range(i, len(self.actions[index])):
			button, action, l, b, clearb = self.actions[index][j]
			grActions.attach(l, 0, j + 1, 1, 1)
			grActions.attach(b, 1, j + 1, 1, 1)
			grActions.attach(clearb, 2, j + 1, 1, 1)
		# Regenereate combobox with removed button added back to it
		# - Store acive item from in combobox
		active, i, index = None, 0, -1
		try:
			active = model.get_value(cbButtonChooser.get_active_iter(), 0)
		except:
			pass
		# Clear entire combobox
		model.clear()
		# Fill it again
		for button, text in self.BUTTONS:
			model.append((None if button is None else nameof(button), text))
			if button is not None:
				if nameof(button) == active:
					index = i
			i += 1
		# Reselect formely active item
		if index >= 0:
			cbButtonChooser.set_active(index)

	def _choose_editor(self, action, cb):
		from scc.gui.ring_editor import RingEditor  # Cannot be imported @ top

		if isinstance(action, Macro):
			from scc.gui.macro_editor import MacroEditor  # Cannot be imported @ top

			e = MacroEditor(self.app, cb)
			e.set_title(_("Edit Macro"))
		elif RingEditor.is_ring_action(action):
			e = RingEditor(self.app, cb)
		else:
			from scc.gui.action_editor import ActionEditor  # Cannot be imported @ top

			e = ActionEditor(self.app, cb)
			e.set_title(_("Edit Action"))
			e.hide_modeshift()
		return e

	def on_actionb_clicked(self, trash, index, clicked_button):
		for i in self.actions[index]:
			button, action, l, b, clearb = i
			if button == clicked_button:

				def on_chosen(id, action):
					b.set_label(action.describe(self.mode))
					i[1] = action

				ae = self._choose_editor(action, on_chosen)
				ae.set_input(self.id, action, mode=self.mode)
				ae.show(self.window)
				return

	def on_ntbMore_switch_page(self, ntb, box, index):
		self.current_page = index
		self._fill_button_chooser()
		self.builder.get_object("cbButtonChooser").set_sensitive(box.get_sensitive())
		self.builder.get_object("btAddAction").set_sensitive(box.get_sensitive())

	def on_nomodbt_clicked(self, button, *a):
		actionButton = self.action_widgets[self.current_page][1]

		def on_chosen(id, action):
			actionButton.set_label(action.describe(self.mode))
			self.nomods[self.current_page] = action

		ae = self._choose_editor(self.nomods[self.current_page], on_chosen)
		ae.set_input(self.id, self.nomods[self.current_page], mode=self.mode)
		ae.show(self.window)

	def on_nomodclear_clicked(self, button, *a):
		self.nomods[self.current_page] = NoAction()
		actionButton = self.action_widgets[self.current_page][1]
		actionButton.set_label(self.nomods[self.current_page].describe(self.mode))

	def on_btAddAction_clicked(self, *a):
		cbButtonChooser = self.builder.get_object("cbButtonChooser")
		item = cbButtonChooser.get_model().get_value(cbButtonChooser.get_active_iter(), 0)
		if item.startswith("Soft"):
			b = getattr(SCButtons, item.split(" ")[-1])
			rng = RangeOP(b, ">=", -1)
			self._add_action(self.current_page, rng, NoAction())
		else:
			b = getattr(SCButtons, item)
			self._add_action(self.current_page, b, NoAction())

	def on_sclSoftLevel_format_value(self, scale, value):
		return "%s%%" % (int(value * 100.0),)

	def on_btClear_clicked(self, *a):
		"""Handler for clear button"""
		action = NoAction()
		if self.ac_callback is not None:
			self.ac_callback(self.id, action)
		self.close()

	def on_btCustomActionEditor_clicked(self, *a):
		"""Handler for 'Custom Editor' button"""
		from scc.gui.action_editor import ActionEditor  # Can't be imported on top

		e = ActionEditor(self.app, self.ac_callback)
		e.set_input(self.id, self._make_action(), mode=self.mode)
		e.hide_action_buttons()
		e.hide_advanced_settings()
		e.set_title(self.window.get_title())
		e.force_page(e.load_component("custom"), True)
		self.send_added_widget(e)
		self.close()
		e.show(self.get_transient_for())

	def on_cbHoldFeedback_toggled(self, cb, *a):
		rvHoldFeedbackAmplitude = self.builder.get_object("rvHoldFeedbackAmplitude")
		rvHoldFeedbackAmplitude.set_reveal_child(cb.get_active())

	def on_btOK_clicked(self, *a):
		"""Handler for OK button"""
		if self.ac_callback is not None:
			self.ac_callback(self.id, self._make_action())
		self.close()

	def _make_action(self):
		"""Generates and returns Action instance"""
		cbHoldFeedback = self.builder.get_object("cbHoldFeedback")
		sclHoldFeedback = self.builder.get_object("sclHoldFeedback")
		normalaction = self._save_modemod(0)
		holdaction = self._save_modemod(1)
		dblaction = self._save_modemod(2)
		if dblaction:
			action = DoubleclickModifier(dblaction, normalaction)
			action.holdaction = holdaction
		elif holdaction:
			action = HoldModifier(holdaction, normalaction)
		else:
			action = normalaction
		action.timeout = self.builder.get_object("adjTime").get_value()

		if cbHoldFeedback.get_active():
			action = FeedbackModifier(HapticPos.BOTH, sclHoldFeedback.get_value(), action)

		return action

	def _save_modemod(self, index):
		"""Generates ModeModifier from page in Notebook"""
		adjSoftLevel = self.builder.get_object("adjSoftLevel")
		pars = []
		for item, action, l, b, clearb in self.actions[index]:
			if isinstance(item, RangeOP) and item.value == -1:
				# Value taken from slider
				item.value = adjSoftLevel.get_value()
			pars += [item, action]
		if self.nomods[index]:
			pars += [self.nomods[index]]
		action = ModeModifier(*pars)
		if len(pars) == 0:
			# No action is actually set
			action = NoAction()
		elif len(pars) == 1:
			# Only default action left
			action = self.nomods[index]
		return action

	def _load_modemod(self, index, action):
		for key in action.mods:
			self._add_action(index, key, action.mods[key])

	def _set_nomod_button(self, index, action):
		if isinstance(action, ModeModifier):
			self._load_modemod(index, action)
			self.nomods[index] = action.default
		else:
			self.nomods[index] = action
		actionButton = self.action_widgets[index][1]
		actionButton.set_label(self.nomods[index].describe(self.mode))

	def allow_first_page(self):
		"""For compatibility with action editor. Does nothing"""

	def set_input(self, id, action, mode=None):
		btDefault = self.builder.get_object("btDefault")
		lblPressAlone = self.builder.get_object("lblPressAlone")
		cbHoldFeedback = self.builder.get_object("cbHoldFeedback")
		sclHoldFeedback = self.builder.get_object("sclHoldFeedback")

		self.id = id
		self._fill_button_chooser()

		if id in STICKS:
			lblPressAlone.set_label(_("(no button pressed)"))
			self.mode = mode = mode or Action.AC_STICK
		elif id in PADS:
			lblPressAlone.set_label(_("(no button pressed)"))
			self.mode = mode = mode or Action.AC_PAD
		else:
			lblPressAlone.set_label(_("(pressed alone)"))
			self.mode = mode = mode or Action.AC_BUTTON

		self.set_title("Modeshift for %s" % (nameof(id) if id in SCButtons.__members__.values() else str(id),))

		if isinstance(action, FeedbackModifier):
			cbHoldFeedback.set_active(True)
			sclHoldFeedback.set_value(action.haptic.get_amplitude())
			action = action.action
		else:
			cbHoldFeedback.set_active(False)
			sclHoldFeedback.set_value(512)

		if isinstance(action, ModeModifier):
			self._load_modemod(0, action)
			self._set_nomod_button(0, action.default)
			self._set_nomod_button(1, NoAction())
			self._set_nomod_button(2, NoAction())
		elif isinstance(action, DoubleclickModifier):  # includes HoldModifier
			self._set_nomod_button(0, action.normalaction)
			self._set_nomod_button(1, action.holdaction)
			self._set_nomod_button(2, action.action)
		self.builder.get_object("adjTime").set_value(action.timeout)

		if mode == Action.AC_OSK:
			# This is kinda bad, but allowing Custom Editor
			# from OSK editor is in TODO
			self.builder.get_object("btCustomActionEditor").set_visible(False)
		if mode != Action.AC_BUTTON:
			for w in ("vbHold", "vbDoubleClick", "lblHold", "lblDoubleClick"):
				self.builder.get_object(w).set_sensitive(False)



"""Syncthing-GTK - StatusIcon."""

import logging
import os

from gi.repository import GLib, GObject, Gtk

from scc.gui.dwsnc import IS_GNOME, IS_UNITY
from scc.tools import _  # gettext function

log = logging.getLogger("StatusIcon")

# Taken from Syncthing-GTK, but got all KDE stuff removed, since it doesn't
# work in latest KDE anymore.


#                | MATE      | Unity      | Cinnamon   | Cairo-Dock (classic) | Cairo-Dock (modern) |
# ----------------+-----------+------------+------------+----------------------+---------------------+
# StatusIconAppI | none      | excellent  | none       | none                 | excellent           |
# StatusIconGTK3 | excellent | none       | very good¹ | very good¹           | none                |
#
# Notes:
#  - StatusIconAppIndicator does not implement any fallback (but the original libappindicator did)
#  - Markers:
#     ¹ Icon cropped


class StatusIcon(GObject.GObject):
	"""Base class for all status icon backends."""

	TRAY_TITLE = _("SC Controller")

	__gsignals__ = {
		"clicked": (GObject.SignalFlags.RUN_FIRST, None, ()),
	}

	__gproperties__ = {
		"active": (
			GObject.TYPE_BOOLEAN,
			"is the icon user-visible?",
			"does the icon back-end think that anything is might be shown to the user?",
			True,
			GObject.PARAM_READWRITE if hasattr(GObject, "PARAM_READWRITE") else GObject.ParamFlags.READWRITE,
		),
	}

	def __init__(self, icon_path, popupmenu, force=False):
		GObject.GObject.__init__(self)
		self.__icon_path = os.path.normpath(os.path.abspath(icon_path))
		self.__popupmenu = popupmenu
		self.__active = True
		self.__visible = False
		self.__hidden = False
		self.__icon = "scc-unknown"
		self.__text = ""
		self.__force = force

	def get_active(self):
		"""Return whether there is at least a chance that the icon might be shown to the user

		If this returns `False` then the icon will definetely not be shown, but if it returns `True` it doesn't have to
		be visible...

		<em>Note:</em> This value is not directly influenced by calling `hide()` and `show()`.

		@return {bool}
		"""
		return self.get_property("active")

	def set(self, icon=None, text=None):
		"""Set the status icon image and descriptive text

		If either of these are `None` their previous value will be used.

		@param {String} icon
		       The name of the icon to show (i.e. `si-syncthing-idle`)
		@param {String} text
		       Some text that indicates what the application is currently doing (generally this be used for the tooltip)
		"""
		if not icon.endswith("-0"):  # si-syncthing-0
			# Ignore first syncing icon state to prevent the icon from flickering
			# into the main notification bar during initialization
			self.__visible = True

		if self.__hidden:
			self._set_visible(False)
		else:
			self._set_visible(self.__visible)

	def hide(self):
		"""Hide the icon

		This method tries its best to ensure the icon is hidden, but there are no guarantees as to how use well its
		going to work.
		"""
		self.__hidden = True
		self._set_visible(False)

	def show(self):
		"""Show a previously hidden icon

		This method tries its best to ensure the icon is hidden, but there are no guarantees as to how use well its
		going to work.
		"""
		self.__hidden = False
		self._set_visible(self.__visible)

	def is_clickable(self) -> bool:
		"""Basically, returns False is appindicator is used."""
		return True

	def _is_forced(self):
		return self.__force

	def _on_click(self, *a):
		self.emit("clicked")

	def _get_icon(self, icon=None):
		"""@internal

		Use `set()` instead.
		"""
		if icon:
			self.__icon = icon
		return self.__icon

	def _get_text(self, text=None):
		"""@internal

		Use `set()` instead.
		"""
		if text:
			self.__text = text
		return self.__text

	def _get_popupmenu(self):
		"""@internal
		"""
		return self.__popupmenu

	def _set_visible(self, visible):
		"""@internal
		"""

	def do_get_property(self, property):
		if property.name == "active":
			return self.__active
		raise AttributeError("Unknown property %s" % property.name)

	def do_set_property(self, property, value):
		if property.name == "active":
			self.__active = value
		else:
			raise AttributeError("unknown property %s" % property.name)


class StatusIconDummy(StatusIcon):
	"""Dummy status icon implementation that does nothing
	"""

	def __init__(self, *args, **kwargs):
		StatusIcon.__init__(self, *args, **kwargs)

		# Pretty unlikely that this will be visible...
		self.set_property("active", False)

	def set(self, icon=None, text=None):
		StatusIcon.set(self, icon, text)

		self._get_icon(icon)
		self._get_text(text)


class StatusIconGTK3(StatusIcon):
	"""Gtk.StatusIcon based status icon backend
	"""

	def __init__(self, *args, **kwargs):
		StatusIcon.__init__(self, *args, **kwargs)

		if not self._is_forced():
			if IS_UNITY:
				# Unity fakes SysTray support but actually hides all icons...
				raise NotImplementedError
			if IS_GNOME:
				# Gnome got broken as well
				raise NotImplementedError

		self._tray = Gtk.StatusIcon()

		self._tray.connect("activate", self._on_click)
		self._tray.connect("popup-menu", self._on_rclick)
		self._tray.connect("notify::embedded", self._on_embedded_change)

		self._tray.set_visible(True)
		self._tray.set_name("sc-controller")
		self._tray.set_title(self.TRAY_TITLE)

		# self._tray.is_embedded() must be called asynchronously
		# See: http://stackoverflow.com/a/6365904/277882
		GLib.idle_add(self._on_embedded_change)

	def destroy(self):
		self.hide()
		self._tray = None

	def set(self, icon=None, text=None):
		StatusIcon.set(self, icon, text)

		self._tray.set_from_icon_name(self._get_icon(icon))
		self._tray.set_tooltip_text(self._get_text(text))

	def _on_embedded_change(self, *args):
		# Without an icon update at this point GTK might consider the icon embedded and visible even through
		# it can't actually be seen...
		self._tray.set_from_file(self._get_icon())

		# An invisible tray icon will never be embedded but it also should not be replaced
		# by a fallback icon
		is_embedded = self._tray.is_embedded() or not self._tray.get_visible()
		if is_embedded != self.get_property("active"):
			self.set_property("active", is_embedded)

	def _on_rclick(self, si, button, time):
		self._get_popupmenu().popup(None, None, None, None, button, time)

	def _set_visible(self, active):
		StatusIcon._set_visible(self, active)

		self._tray.set_visible(active)


class StatusIconDBus(StatusIcon):
	pass


class StatusIconAppIndicator(StatusIconDBus):
	"""Unity's AppIndicator3.Indicator based status icon backend
	"""

	def __init__(self, *args, **kwargs):
		StatusIcon.__init__(self, *args, **kwargs)

		try:
			import gi

			try:
				gi.require_version("AyatanaAppIndicator3", "0.1")
				from gi.repository import AyatanaAppIndicator3 as appindicator
			except ImportError:
				log.warning(
					"Failed to import AyatanaAppIndicator3, trying fallback to an old implementation of AppIndicator3!",
				)
				gi.require_version("AppIndicator3", "0.1")
				from gi.repository import AppIndicator3 as appindicator

			self._status_active = appindicator.IndicatorStatus.ACTIVE
			self._status_passive = appindicator.IndicatorStatus.PASSIVE
		except ImportError:
			log.warning("Failed to import AppIndicator3")
			raise NotImplementedError

		category = appindicator.IndicatorCategory.APPLICATION_STATUS
		# Whatever icon is set here will be used as a tooltip icon during the entire time to icon is shown
		self._tray = appindicator.Indicator.new("sc-controller", self._get_icon(), category)
		self._tray.set_menu(self._get_popupmenu())
		self._tray.set_title(self.TRAY_TITLE)

	def _set_visible(self, active):
		StatusIcon._set_visible(self, active)

		self._tray.set_status(self._status_active if active else self._status_passive)

	def is_clickable(self) -> bool:
		return False

	def destroy(self):
		self.hide()
		self._tray = None

	def set(self, icon=None, text=None):
		StatusIcon.set(self, icon, text)

		self._tray.set_icon_full(self._get_icon(icon), self._get_text(text))


class StatusIconProxy(StatusIcon):
	def __init__(self, *args, **kwargs):
		StatusIcon.__init__(self, *args, **kwargs)

		self._arguments = (args, kwargs)
		self._status_fb = None
		self._status_gtk = None
		self.set("scc-unknown", "")

		# Do not ever force-show indicators when they do not think they'll work
		if "force" in self._arguments[1]:
			del self._arguments[1]["force"]

		try:
			# Try loading GTK native status icon
			self._status_gtk = StatusIconGTK3(*args, **kwargs)
			self._status_gtk.connect("clicked", self._on_click)
			self._status_gtk.connect("notify::active", self._on_notify_active_gtk)
			self._on_notify_active_gtk()

			log.info("Using backend StatusIconGTK3 (primary)")
		except NotImplementedError:
			# Directly load fallback implementation
			self._load_fallback()

	def _on_click(self, *args):
		self.emit("clicked")

	def _on_notify_active_gtk(self, *args):
		if self._status_fb:
			# Hide fallback icon if GTK icon is active and vice-versa
			if self._status_gtk.get_active():
				self._status_fb.hide()
			else:
				self._status_fb.show()
		elif not self._status_gtk.get_active():
			# Load fallback implementation
			self._load_fallback()

	def _on_notify_active_fb(self, *args):
		active = False
		if self._status_gtk and self._status_gtk.get_active():
			active = True
		if self._status_fb and self._status_fb.get_active():
			active = True
		self.set_property("active", active)

	def _load_fallback(self):
		status_icon_backends = [StatusIconAppIndicator, StatusIconDummy]

		if not self._status_fb:
			for StatusIconBackend in status_icon_backends:
				try:
					self._status_fb = StatusIconBackend(*self._arguments[0], **self._arguments[1])
					self._status_fb.connect("clicked", self._on_click)
					self._status_fb.connect("notify::active", self._on_notify_active_fb)
					self._on_notify_active_fb()

					log.warning("StatusIcon: Using backend %s (fallback)" % StatusIconBackend.__name__)
					break
				except NotImplementedError:
					continue

			# At least the dummy backend should have been loaded at this point...
			assert self._status_fb

		# Update fallback icon
		self.set(self._icon, self._text)

	def is_clickable(self) -> bool:
		if self._status_gtk:
			return self._status_gtk.is_clickable()
		if self._status_fb:
			return self._status_fb.is_clickable()
		return False

	def set(self, icon=None, text=None):
		self._icon = icon
		self._text = text

		if self._status_gtk:
			self._status_gtk.set(icon, text)
		if self._status_fb:
			self._status_fb.set(icon, text)

	def hide(self):
		if self._status_gtk:
			self._status_gtk.hide()
		if self._status_fb:
			self._status_fb.hide()

	def destroy(self):
		if self._status_gtk:
			self._status_gtk.destroy()
		if self._status_fb:
			self._status_fb.destroy()

	def show(self):
		if self._status_gtk:
			self._status_gtk.show()
		if self._status_fb:
			self._status_fb.show()


def get_status_icon(*args, **kwargs):
	# Try selecting backend based on environment variable
	if "STATUS_BACKEND" in os.environ:
		kwargs["force"] = True

		status_icon_backend_name = "StatusIcon%s" % (os.environ.get("STATUS_BACKEND"))
		if status_icon_backend_name in globals():
			try:
				status_icon = globals()[status_icon_backend_name](*args, **kwargs)
				log.info("StatusIcon: Using requested backend %s" % (status_icon_backend_name))
				return status_icon
			except NotImplementedError:
				log.error("StatusIcon: Requested backend %s is not supported" % (status_icon_backend_name))
		else:
			log.error("StatusIcon: Requested backend %s does not exist" % (status_icon_backend_name))

		return StatusIconDummy(*args, **kwargs)

	# Use proxy backend to determine the correct backend while the application is running
	return StatusIconProxy(*args, **kwargs)



#!/usr/bin/env python3
"""SC Controller - remotepad driver

This is implementation or protocol used by Retroarch's Remote RetroPad core.
Based on https://github.com/libretro/RetroArch/blob/master/cores/libretro-net-retropad.
"""

import ctypes
import logging
import socket
from ctypes import CFUNCTYPE, POINTER, byref, c_void_p, cast

from scc.constants import ControllerFlags
from scc.controller import Controller
from scc.tools import find_library

log = logging.getLogger("remotepad")


class ControllerInput(ctypes.Structure):
	_fields_ = [
		("buttons", ctypes.c_uint32),
		("ltrig", ctypes.c_uint8),
		("rtrig", ctypes.c_uint8),
		("stick_x", ctypes.c_int16),
		("stick_y", ctypes.c_int16),
		("lpad_x", ctypes.c_int16),
		("lpad_y", ctypes.c_int16),
		("rpad_x", ctypes.c_int16),
		("rpad_y", ctypes.c_int16),
		("cpad_x", ctypes.c_int16),
		("cpad_y", ctypes.c_int16),
		("gpitch", ctypes.c_int16),
		("groll", ctypes.c_int16),
		("gyaw", ctypes.c_int16),
		("q1", ctypes.c_int16),
		("q2", ctypes.c_int16),
		("q3", ctypes.c_int16),
		("q4", ctypes.c_int16),
	]


MapperInputCB = CFUNCTYPE(None, c_void_p, POINTER(ControllerInput))


class Mapper(ctypes.Structure):
	_fields_ = [
		("input", MapperInputCB),
	]


class RemotePad(ctypes.Structure):
	_fields_ = [
		("mapper", POINTER(Mapper)),
		("input", ControllerInput),
	]


class RemoteJoypadMessage(ctypes.Structure):
	_fields_ = [
		("port", ctypes.c_int),
		("device", ctypes.c_int),
		("index", ctypes.c_int),
		("id", ctypes.c_int),
		("state", ctypes.c_uint16),
	]


class RemotePadController(Controller):
	flags = (
		ControllerFlags.HAS_DPAD
		| ControllerFlags.NO_GRIPS
		| ControllerFlags.HAS_RSTICK
		| ControllerFlags.SEPARATE_STICK
	)

	def __init__(self, driver, address):
		Controller.__init__(self)
		self._id = "rpad%s" % (self._id,)
		self._driver = driver
		self._address = address
		self._enabled = True
		self._mapper = Mapper()
		self._mapper.input = MapperInputCB(self._input)
		self._old_state = ControllerInput()
		self._state_size = ctypes.sizeof(ControllerInput)
		self._pad = RemotePad()
		self._pad.mapper = POINTER(Mapper)(self._mapper)

	def get_type(self):
		return "rpad"

	def _remove(self, *a):
		self._driver._remove(self._address)

	def turnoff(self):
		log.debug("Disconnecting %s", self._address)
		self._disabled = True
		self._driver.daemon.remove_controller(self)
		self._driver.daemon.get_scheduler().schedule(10.0, self._remove)

	def _input(self, trash, data):
		if self._enabled and self.mapper:
			self.mapper.input(self, self._old_state, data.contents)
			ctypes.memmove(byref(self._old_state), data, self._state_size)

	def get_gui_config_file(self):
		return "remotepad.json"


class Driver:
	PORT = 55400

	def __init__(self, daemon, config):
		self._controllers = {}
		self.daemon = daemon
		self.config = config
		self._lib = find_library("libremotepad")
		self._lib.remotepad_input.argtypes = [POINTER(RemotePad), POINTER(RemoteJoypadMessage)]
		self._lib.remotepad_input.restype = None
		self._size = ctypes.sizeof(RemoteJoypadMessage)
		self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		server_address = ("0.0.0.0", self.PORT)
		self.sock.bind(server_address)
		poller = self.daemon.get_poller()
		poller.register(self.sock.fileno(), poller.POLLIN, self.on_data_ready)
		log.info("Listening on %s:%s", *server_address)

	def _remove(self, address):
		if address in self._controllers:
			del self._controllers[address]

	def on_data_ready(self, *a):
		data, source = self.sock.recvfrom(self._size)
		address, port = source
		controller = None
		if address not in self._controllers:
			controller = RemotePadController(self, address)
			self._controllers[address] = controller
			self.daemon.add_controller(controller)
		else:
			controller = self._controllers[address]

		self._lib.remotepad_input(controller._pad, cast(ctypes.c_char_p(data), POINTER(RemoteJoypadMessage)))


def init(daemon, config):
	"""Registers hotplug callback for controller dongle"""
	_drv = Driver(daemon, config)
	return True



#!/usr/bin/env python3
"""SC-Controller - Action Editor - OSK Button Component

Binds controller buttons on on on on on... screen keyboard.
Retuses ButtonsComponent, but hides image, so user can't select mouse or gamepad
button.
"""

import logging

from scc.actions import Action
from scc.gui.ae import AEComponent
from scc.gui.ae.buttons import ButtonsComponent
from scc.tools import _

log = logging.getLogger("AE.Buttons")

__all__ = ["OSKButtonsComponent"]


class OSKButtonsComponent(ButtonsComponent):
	CTXS = Action.AC_OSK
	PRIORITY = 1
	IMAGES = {}

	def get_button_title(self):
		return _("Key")

	def load(self):
		if not self.loaded:
			AEComponent.load(self)
			self.builder.get_object("lblClickAnyButton").set_visible(False)



"""SC Controller - Action Editor

Allows to edit button or trigger action.
"""


import logging
import os

from gi.repository import Gtk

from scc.gui.gdk_to_key import keyevent_to_key
from scc.uinput import Keys

log = logging.getLogger("KeyGrabber")

MODIFIERS = [
	Keys.KEY_LEFTCTRL,
	Keys.KEY_LEFTMETA,
	Keys.KEY_LEFTALT,
	Keys.KEY_RIGHTALT,
	Keys.KEY_RIGHTMETA,
	Keys.KEY_RIGHTCTRL,
	Keys.KEY_LEFTSHIFT,
	Keys.KEY_RIGHTSHIFT,
]


def merge_modifiers(mods):
	return "+".join([key.name.split("_")[-1] for key in mods])


# Just to speed shit up, KeyGrabber is singleton
class KeyGrabber:
	GLADE = "key_grabber.glade"
	_singleton = None

	def __new__(cls, *a):
		if cls._singleton is None:
			cls._singleton = object.__new__(cls)
		return cls._singleton

	def __init__(self, app):
		self.app = app
		self.builder = None
		self.active_mods = []

	def grab(self, modal_for, action, callback):
		if self.builder is None:
			self.setup_widgets()
		self.callback = callback
		self.builder.get_object("lblKey").set_label("...")
		# TODO: Display 'action' in lblKey if possible
		for key in MODIFIERS:
			self.builder.get_object("tg" + key.name).set_active(False)
		self.active_mods = []
		self.window.set_transient_for(modal_for)
		self.window.set_modal(True)
		self.window.show()
		self.window.set_focus()

	def setup_widgets(self):
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(self.app.gladepath, self.GLADE))
		self.window = self.builder.get_object("KeyGrab")
		self.builder.connect_signals(self)

	def on_KeyGrab_destroy(self, *a):
		# Don't allow destroying
		return True

	def on_keyGrab_key_press_event(self, trash, event):
		"""Handle keypress on "Grab Key" dialog.

		Remembers modifiers and displays text in middle of dialog.
		Dialog is dismissed (and key is accepted) by key_release handler bellow.
		"""
		key = keyevent_to_key(event)
		if key is None:
			log.warning("Unknown keycode %s/%s" % (event.keyval, event.hardware_keycode))
			return

		if key in MODIFIERS:
			self.active_mods.append(key)
			self.builder.get_object("tg" + key.name).set_active(True)
			self.builder.get_object("lblKey").set_label(merge_modifiers(self.active_mods))
			return

		label = merge_modifiers(self.active_mods)
		if len(self.active_mods) > 0:
			label = label + "+" + key.name
		else:
			label = key.name
		self.builder.get_object("lblKey").set_label(label)

	def on_keyGrab_key_release_event(self, trash, event):
		"""Handles keyrelease on "Grab Key" dialog.

		Key is accepted if either:
		- released key is not modifier
		- released key is modifier, but there is no other modifier key pressed

		Calls callback if key is accepted
		"""
		key = keyevent_to_key(event)
		if key is not None:
			if key in MODIFIERS:
				if key in self.active_mods:
					if len(self.active_mods) == 1:
						# Releasing last modifier
						self.callback([key])
						self.window.hide()
						return
					self.active_mods.remove(key)
					self.builder.get_object("tg" + key.name).set_active(False)
				self.builder.get_object("lblKey").set_label(
					"+".join([key.name.split("_")[-1] for key in self.active_mods]),
				)
				return

			self.callback(self.active_mods + [key])
			self.window.hide()

	def on_tgkey_toggled(self, obj, *a):
		"""Handles when user clicks on modifier buttons in "Grab Key" dialog
		"""
		for key in MODIFIERS:
			if self.builder.get_object("tg" + key.name) == obj:
				if obj.get_active() and key not in self.active_mods:
					self.active_mods.append(key)
					self.builder.get_object("lblKey").set_label(merge_modifiers(self.active_mods))
				elif not obj.get_active() and key in self.active_mods:
					self.active_mods.remove(key)
					self.builder.get_object("lblKey").set_label(merge_modifiers(self.active_mods))
				return



"""SC Controller - On Screen Keyboard Actions.

Special Actions that are used to bind functions like closing keyboard or moving
cursors around.

Actions defined here are *not* automatically registered, but OSD Keyboard
and its binding editor enables them to use with 'OSK.something'
syntax.
"""

import logging

from scc.actions import Action, SpecialAction
from scc.constants import LEFT, RIGHT, TRIGGER_HALF

log = logging.getLogger("OSDKeyActs")
_ = lambda x: x


class OSKAction(Action, SpecialAction):
	def __init__(self, *a):
		Action.__init__(self, *a)
		self.speed = 1.0

	def set_speed(self, x, y, z):
		self.speed = x
		return True

	def trigger(self, mapper, p, old_p):
		if p * self.speed >= TRIGGER_HALF and old_p * self.speed < TRIGGER_HALF:
			self.button_press(mapper)
		elif p * self.speed < TRIGGER_HALF and old_p * self.speed >= TRIGGER_HALF:
			self.button_release(mapper)


class CloseOSKAction(OSKAction):
	SA: str = "close"
	COMMAND: str = SA

	def describe(self, context):
		if context == Action.AC_OSK:
			return _("Hide")
		return _("Hide Keyboard")

	def to_string(self, multiline=False, pad=0):
		return (" " * pad) + "OSK.%s()" % (self.COMMAND,)

	def button_press(self, mapper):
		self.execute(mapper)

	def button_release(self, mapper):
		pass


class OSKCursorAction(Action, SpecialAction):
	SA: str = "cursor"
	COMMAND: str = SA

	def __init__(self, side):
		Action.__init__(self, side)
		if hasattr(side, "name"):
			side = side.name
		self.speed = (1.0, 1.0)
		self.side = side

	def set_speed(self, x, y, z):
		self.speed = (x, y)
		return True

	def whole(self, mapper, x, y, what):
		self.execute(mapper, x, y)

	def describe(self, context):
		if self.side == LEFT:
			return _("Move LEFT Cursor")
		if self.side == RIGHT:
			return _("Move RIGHT Cursor")
		return _("Move Cursor")

	def to_string(self, multiline=False, pad=0):
		return (" " * pad) + "OSK.%s(%s)" % (self.COMMAND, self.side)


class MoveOSKAction(OSKAction):
	SA: str = "move"
	COMMAND: str = SA

	def whole(self, mapper, x, y, what):
		self.execute(mapper, x, y)

	def describe(self, context):
		return _("Move Keyboard")

	def to_string(self, multiline=False, pad=0):
		return (" " * pad) + "OSK.%s()" % (self.COMMAND,)


class OSKPressAction(OSKAction):
	SA: str = "press"
	COMMAND: str = SA

	def __init__(self, side):
		OSKAction.__init__(self, side)
		if hasattr(side, "name"):
			side = side.name
		self.side = side

	def describe(self, context):
		if context == Action.AC_OSK:
			return _("Press Key")
		if self.side == LEFT:
			return _("Press Key Under LEFT Cursor")
		return _("Press Key Under RIGHT Cursor")

	def button_press(self, mapper):
		self.execute(mapper, True)

	def button_release(self, mapper):
		self.execute(mapper, False)

	def to_string(self, multiline=False, pad=0):
		return (" " * pad) + "OSK.%s(%s)" % (self.COMMAND, self.side)



#!/usr/bin/env python3
"""SC-Controller goes through all modules in scc.drivers package and calls
init(daemon) methods from every module that defines it.

Drivers then can use daemon.add_mainloop() method to add code that should
be run in every mainloop iteration and daemon.add_controller() to add and
daemon.remove_controller() to remove Controller instances.

Additionaly, start(daemon) method is called from each module that defines it
just before daemon startup is complete.

Assigning Mapper to Controller is handled by daemon.
"""

MOD_INIT_ORDER = (
	# Modules mentioned here are initialized before everything else, in this exact order.
	"scc.drivers.usb",
	"scc.drivers.evdevdrv",
	"scc.drivers.hiddrv",
)



"""VDF file reader.

Copyright (C) 2017 Kozec

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 as published by
the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import vdf


def parse_vdf(file):
	return vdf.parse(file, mapper=vdf.VDFDict, merge_duplicate_keys=False)


def ensure_list(value):
	"""If value is list, returns same value.

	Otherwise, returns [ value ]
	"""
	return value if type(value) == list else [value]


if __name__ == "__main__":
	print(parse_vdf(open("test.vdf")))



#!/usr/bin/env python3
"""SC-Controller - Quick OSD Menu

Controled by buttons instead of stick. Fast to use, but can display only
limited number of items
"""

import logging
import os
import sys

from gi.repository import GLib, Gtk

from scc.config import Config
from scc.menu_data import MenuItem, Submenu
from scc.osd import OSDWindow
from scc.osd.menu import Menu, MenuIcon
from scc.paths import get_share_path
from scc.tools import _, find_icon, find_menu

log = logging.getLogger("osd.quickmenu")


class QuickMenu(Menu):
	BUTTONS = ["A", "B", "X", "Y", "LB", "RB"]
	BUTTON_INDEXES = [0, 1, 2, 3, 7, 8]  # indexes to gui->buttons list
	# in controller gui config

	def __init__(self, cls="osd-menu"):
		Menu.__init__(self, cls)
		self._cancel_with = "START"
		self._pressed = []
		self._icons = []
		self._timer = None

	def generate_widget(self, item):
		"""In QuickMenu, everything but submenus and simple
		menuitems is ignored.
		"""
		if self._button_index >= len(self.BUTTONS):
			return None
		if isinstance(item, (MenuItem, Submenu)):
			widget = Gtk.Button.new_with_label(item.label)
			widget.set_relief(Gtk.ReliefStyle.NONE)
			if hasattr(widget.get_children()[0], "set_xalign"):
				widget.get_children()[0].set_xalign(0)
			else:
				widget.get_children()[0].set_halign(Gtk.Align.START)
			if isinstance(item, Submenu):
				item.callback = self.show_submenu
				label1 = widget.get_children()[0]
				label2 = Gtk.Label(_(">>"))
				label2.set_property("margin-left", 30)
				box = Gtk.Box(Gtk.Orientation.HORIZONTAL)
				widget.remove(label1)
				box.pack_start(label1, True, True, 1)
				box.pack_start(label2, False, True, 1)
				widget.add(box)
				widget.set_name("osd-menu-item")
			elif item.id is None:
				# Ignored as well
				return None
			else:
				widget.set_name("osd-menu-item")

			item.button = self.BUTTONS[self._button_index]
			self._button_index += 1

			icon_file, has_colors = find_icon("buttons/%s" % item.button, False)
			icon = MenuIcon(icon_file, has_colors)
			label = widget.get_children()[0]
			for c in [] + widget.get_children():
				widget.remove(c)
			self._icons.append(icon)
			box = Gtk.Box()
			box.pack_start(icon, False, True, 0)
			box.pack_start(label, True, True, 10)
			widget.add(box)
			return widget
		return None

	def _add_arguments(self):
		OSDWindow._add_arguments(self)
		self.argparser.add_argument(
			"--cancel-with",
			type=str,
			metavar="button",
			default="START",
			help="button used to cancel menu (default: START)",
		)
		self.argparser.add_argument(
			"--timeout", type=int, default=5, help="how many seconds before menu is automatically canceled",
		)
		self.argparser.add_argument(
			"--cancel-with-release", action="store_true", help="cancel menu with button release instead of button press",
		)
		self.argparser.add_argument(
			"--from-profile", "-p", type=str, metavar="profile_file menu_name", help="load menu items from profile file",
		)
		self.argparser.add_argument(
			"--from-file", "-f", type=str, metavar="filename", help="load menu items from json file",
		)
		self.argparser.add_argument("--print-items", action="store_true", help="prints menu items to stdout")
		self.argparser.add_argument("items", type=str, nargs="*", metavar="id title", help="Menu items")

	def _check_on_screen_position(self, quick=False):
		pass

	def lock_inputs(self):
		def success(*a):
			log.error("Sucessfully locked input")
			config = self.controller.load_gui_config(os.path.join(get_share_path(), "images"))
			if config and config["gui"] and config["gui"]["buttons"]:
				buttons = config["gui"]["buttons"]
				try:
					for i in range(len(self._icons)):
						icon = self._icons[i]
						name = buttons[self.BUTTON_INDEXES[i]]
						filename, trash = find_icon("buttons/%s" % name)
						icon.set_filename(filename)
						icon.queue_draw()
				except IndexError:
					pass

		locks = [x for x in self.BUTTONS] + [self._cancel_with]
		self.controller.lock(success, self.on_failed_to_lock, *locks)

	def parse_argumets(self, argv):
		if not OSDWindow.parse_argumets(self, argv):
			print("failed to parse args")
			return False
		if not self.parse_menu():
			print("failed to parse menu")
			return False
		if not self.config:
			self.config = Config()

		self._cancel_with = self.args.cancel_with
		self._timeout = self.args.timeout

		# Create buttons that are displayed on screen
		items = self.items.generate(self)
		self.items = []
		self._button_index = 0
		for item in items:
			item.widget = self.generate_widget(item)
			if item.widget is not None:
				self.items.append(item)
		self.pack_items(self.parent, self.items)
		if len(self.items) == 0:
			print("%s: error: no items in menu" % (sys.argv[0]), file=sys.stderr)
			return False

		return True

	def next_item(self, direction):
		pass

	def select(self, index):
		pass

	def show_submenu(self, trash, trash2, trash3, menuitem):
		"""Called when user chooses menu item pointing to submenu"""
		filename = find_menu(menuitem.filename)
		if filename:
			self._submenu = QuickMenu()
			sub_pos = list(self.position)
			for i in (0, 1):
				sub_pos[i] = sub_pos[i] - self.SUBMENU_OFFSET if sub_pos[i] < 0 else sub_pos[i] + self.SUBMENU_OFFSET

			self._submenu.use_config(self.config)
			self._submenu.parse_argumets(
				[
					"menu.py",
					"-x",
					str(sub_pos[0]),
					"-y",
					str(sub_pos[1]),
					"--timeout",
					str(self._timeout),
					"--from-file",
					filename,
				],
			)
			self._submenu.set_is_submenu()
			self._submenu.use_daemon(self.daemon)
			self._submenu.connect("destroy", self.on_submenu_closed)
			self._submenu.controller = self.controller
			self._submenu.show()
			self.cancel_timer()

	def on_submenu_closed(self, *a):
		# Quickmenu can have submenus, but everything cancels at once when
		# last Quickmenu in hierarchy is canceled or timeouts
		if self._submenu.get_exit_code() in (0, -2):
			self._menuid = self._submenu._menuid
		self._selected = self._submenu._selected
		self.quit(self._submenu.get_exit_code())

	def pressed(self, what):
		"""Called when button is pressed. If menu with that button assigned
		exists, it is hilighted.
		"""
		for item in self.items:
			if item.button == what:
				self._pressed.append(item)
				item.widget.set_name("osd-menu-item-selected")

	def released(self, what):
		"""Called when button is pressed. If menu with that button assigned
		exists, it is hilighted.
		"""
		last = None
		for item in self.items:
			if item.button == what:
				while item in self._pressed:
					self._pressed.remove(item)
				item.widget.set_name("osd-menu-item")
				last = item

		if len(self._pressed) == 0 and last is not None:
			if last.callback:
				last.callback(self, self.daemon, self.controller, last)
			else:
				self._selected = last
				self.quit(0)

	def on_timeout(self, *a):
		self.quit(-1)

	def show(self, *a):
		Menu.show(self, *a)
		self.restart_timer()

	def restart_timer(self):
		self.cancel_timer()
		self._timer = GLib.timeout_add_seconds(self._timeout, self.on_timeout)

	def cancel_timer(self):
		if self._timer:
			GLib.source_remove(self._timer)
			self._timer = None

	def on_event(self, daemon, what, data):
		if self._submenu:
			return self._submenu.on_event(daemon, what, data)
		if what == self._cancel_with:
			if data[0] == 0:  # Button released
				self.quit(-1)
		elif what in self.BUTTONS:
			self.restart_timer()
			if data[0] == 1:  # Button pressed
				self.pressed(what)
			else:  # Released
				self.released(what)


if __name__ == "__main__":
	import gi

	gi.require_version("Gtk", "3.0")
	gi.require_version("Rsvg", "2.0")
	gi.require_version("GdkX11", "3.0")

	from scc.paths import get_share_path
	from scc.tools import init_logging

	init_logging()

	m = QuickMenu()
	if not m.parse_argumets(sys.argv):
		sys.exit(1)
	m.run()
	if m.get_exit_code() == 0:
		print(m.get_selected_item_id())
	sys.exit(m.get_exit_code())



#!/usr/bin/env python3
"""SC-Controller - Autoswitch Daemon

Observes active window and commands scc-daemon to change profiles as needed.
"""

import logging
import os
import sys

from scc.x11.autoswitcher import AutoSwitcher

log = logging.getLogger("AutoSwitcher-Daemon")

if __name__ == "__main__":
	from scc.tools import init_logging, set_logging_level

	init_logging(suffix=" AutoSwitcher")
	set_logging_level("debug" in sys.argv, "debug" in sys.argv)

	if "DISPLAY" not in os.environ:
		log.error("DISPLAY env variable not set.")
		sys.exit(1)

	d = AutoSwitcher()
	d.run()
	sys.exit(d.exit_code)



#!/usr/bin/env python3
"""SC-Controller - KEYCODE_TO_KEY

Similar to GDK_TO_KEY, maps X11 keycodes to Keys.KEY_* constants.
Used by OSD keyboard
"""

from .gdk_to_key import KEYCODE_TO_KEY

KEY_TO_KEYCODE = {KEYCODE_TO_KEY[a]: a for a in KEYCODE_TO_KEY}
